# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# ruff: noqa

"""QPlainTextEdit base class"""

# pylint: disable=C0103
# pylint: disable=R0903
# pylint: disable=R0911
# pylint: disable=R0201

import os
import re
import sys
from collections import OrderedDict

from qtpy.QtCore import QEvent, QEventLoop, QPoint, Qt, Signal, Slot
from qtpy.QtGui import (
    QClipboard,
    QColor,
    QMouseEvent,
    QPalette,
    QTextCharFormat,
    QTextCursor,
    QTextFormat,
    QTextOption,
)
from qtpy.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QListWidget,
    QListWidgetItem,
    QMainWindow,
    QPlainTextEdit,
    QTextEdit,
    QToolTip,
)

from guidata import qthelpers as qth
from guidata.config import CONF
from guidata.configtools import get_font, get_icon
from guidata.widgets.console.calltip import CallTipWidget
from guidata.widgets.console.mixins import BaseEditMixin
from guidata.widgets.console.terminal import ANSIEscapeCodeHandler


def insert_text_to(cursor, text, fmt):
    """Helper to print text, taking into account backspaces"""
    while True:
        index = text.find(chr(8))  # backspace
        if index == -1:
            break
        cursor.insertText(text[:index], fmt)
        if cursor.positionInBlock() > 0:
            cursor.deletePreviousChar()
        text = text[index + 1 :]
    cursor.insertText(text, fmt)


class CompletionWidget(QListWidget):
    """Completion list widget"""

    sig_show_completions = Signal(object)

    def __init__(self, parent, ancestor):
        QListWidget.__init__(self, ancestor)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setWindowFlags(Qt.SubWindow | Qt.FramelessWindowHint)
        self.textedit = parent
        self.completion_list = None
        self.case_sensitive = False
        self.enter_select = None
        self.hide()
        self.itemActivated.connect(self.item_selected)

    def setup_appearance(self, size, font):
        """

        :param size:
        :param font:
        """
        self.resize(*size)
        self.setFont(font)

    def show_list(self, completion_list, automatic=True):
        """

        :param completion_list:
        :param automatic:
        :return:
        """
        types = [c[1] for c in completion_list]
        completion_list = [c[0] for c in completion_list]
        if len(completion_list) == 1 and not automatic:
            self.textedit.insert_completion(completion_list[0])
            return

        self.completion_list = completion_list
        self.clear()

        icons_map = {
            "instance": "attribute",
            "statement": "attribute",
            "method": "method",
            "function": "function",
            "class": "class",
            "module": "module",
        }

        self.type_list = types
        if any(types):
            for c, t in zip(completion_list, types):
                icon = icons_map.get(t, "no_match")
                self.addItem(QListWidgetItem(get_icon(icon + ".png"), c))
        else:
            self.addItems(completion_list)

        self.setCurrentRow(0)

        QApplication.processEvents(QEventLoop.ExcludeUserInputEvents)
        self.show()
        self.setFocus()
        self.raise_()

        # Retrieving current screen height
        srect = self.textedit.screen().availableGeometry()
        screen_right = srect.right()
        screen_bottom = srect.bottom()

        point = self.textedit.cursorRect().bottomRight()
        point.setX(point.x() + self.textedit.get_linenumberarea_width())
        point = self.textedit.mapToGlobal(point)

        # Computing completion widget and its parent right positions
        comp_right = point.x() + self.width()
        ancestor = self.parentWidget()
        if ancestor is None:
            anc_right = screen_right
        else:
            anc_right = min([ancestor.x() + ancestor.width(), screen_right])

        # Moving completion widget to the left
        # if there is not enough space to the right
        if comp_right > anc_right:
            point.setX(point.x() - self.width())

        # Computing completion widget and its parent bottom positions
        comp_bottom = point.y() + self.height()
        ancestor = self.parentWidget()
        if ancestor is None:
            anc_bottom = screen_bottom
        else:
            anc_bottom = min([ancestor.y() + ancestor.height(), screen_bottom])

        # Moving completion widget above if there is not enough space below
        x_position = point.x()
        if comp_bottom > anc_bottom:
            point = self.textedit.cursorRect().topRight()
            point = self.textedit.mapToGlobal(point)
            point.setX(x_position)
            point.setY(point.y() - self.height())

        if ancestor is not None:
            # Useful only if we set parent to 'ancestor' in __init__
            point = ancestor.mapFromGlobal(point)
        self.move(point)

        if str(self.textedit.completion_text):
            # When initialized, if completion text is not empty, we need
            # to update the displayed list:
            self.update_current()

        # signal used for testing
        self.sig_show_completions.emit(completion_list)

    def hide(self):
        """ """
        QListWidget.hide(self)
        self.textedit.setFocus()

    def keyPressEvent(self, event):
        """

        :param event:
        """
        text, key = event.text(), event.key()
        alt = event.modifiers() & Qt.AltModifier
        shift = event.modifiers() & Qt.ShiftModifier
        ctrl = event.modifiers() & Qt.ControlModifier
        modifier = shift or ctrl or alt
        if (
            key in (Qt.Key_Return, Qt.Key_Enter) and self.enter_select
        ) or key == Qt.Key_Tab:
            self.item_selected()
        elif key in (
            Qt.Key_Return,
            Qt.Key_Enter,
            Qt.Key_Left,
            Qt.Key_Right,
        ) or text in (".", ":"):
            self.hide()
            self.textedit.keyPressEvent(event)
        elif (
            key
            in (
                Qt.Key_Up,
                Qt.Key_Down,
                Qt.Key_PageUp,
                Qt.Key_PageDown,
                Qt.Key_Home,
                Qt.Key_End,
                Qt.Key_CapsLock,
            )
            and not modifier
        ):
            QListWidget.keyPressEvent(self, event)
        elif len(text) or key == Qt.Key_Backspace:
            self.textedit.keyPressEvent(event)
            self.update_current()
        elif modifier:
            self.textedit.keyPressEvent(event)
        else:
            self.hide()
            QListWidget.keyPressEvent(self, event)

    def update_current(self):
        """ """
        completion_text = str(self.textedit.completion_text)

        if completion_text:
            for row, completion in enumerate(self.completion_list):
                if not self.case_sensitive:
                    print(completion_text)  # spyder: test-skip
                    completion = completion.lower()
                    completion_text = completion_text.lower()
                if completion.startswith(completion_text):
                    self.setCurrentRow(row)
                    self.scrollTo(self.currentIndex(), QAbstractItemView.PositionAtTop)
                    break
            else:
                self.hide()
        else:
            self.hide()

    def focusOutEvent(self, event):
        """

        :param event:
        """
        event.ignore()
        # Don't hide it on Mac when main window loses focus because
        # keyboard input is lost
        # Fixes Issue 1318
        if sys.platform == "darwin":
            if event.reason() != Qt.ActiveWindowFocusReason:
                self.hide()
        else:
            self.hide()

    def item_selected(self, item=None):
        """

        :param item:
        """
        if item is None:
            item = self.currentItem()
        self.textedit.insert_completion(str(item.text()))
        self.hide()


class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
    """Text edit base widget"""

    BRACE_MATCHING_SCOPE = ("sof", "eof")
    cell_separators = None
    focus_in = Signal()
    zoom_in = Signal()
    zoom_out = Signal()
    zoom_reset = Signal()
    focus_changed = Signal()
    sig_eol_chars_changed = Signal(str)

    def __init__(self, parent=None):
        QPlainTextEdit.__init__(self, parent)
        BaseEditMixin.__init__(self)
        self.setAttribute(Qt.WA_DeleteOnClose)

        self.extra_selections_dict = OrderedDict()

        self.textChanged.connect(self.changed)
        self.cursorPositionChanged.connect(self.cursor_position_changed)

        self.indent_chars = " " * 4
        self.tab_stop_width_spaces = 4

        # Code completion / calltips
        if parent is not None:
            mainwin = parent
            while not isinstance(mainwin, QMainWindow):
                mainwin = mainwin.parent()
                if mainwin is None:
                    break
            if mainwin is not None:
                parent = mainwin

        self.completion_widget = CompletionWidget(self, parent)
        self.codecompletion_auto = False
        self.codecompletion_case = True
        self.codecompletion_enter = False
        self.completion_text = ""
        self.setup_completion()

        self.calltip_widget = CallTipWidget(self, hide_timer_on=False)
        self.calltips = True
        self.calltip_position = None

        self.has_cell_separators = False
        self.highlight_current_cell_enabled = False

        # The color values may be overridden by the syntax highlighter
        # Highlight current line color
        self.currentline_color = QColor(Qt.red).lighter(190)
        self.currentcell_color = QColor(Qt.red).lighter(194)

        # Brace matching
        self.bracepos = None
        self.matched_p_color = QColor(Qt.green)
        self.unmatched_p_color = QColor(Qt.red)

        self.last_cursor_cell = None

    def setup_completion(self, size=None, font=None):
        """

        :param size:
        :param font:
        """
        size = size or CONF.get("console", "codecompletion/size")
        font = font or get_font(CONF, "texteditor", "font")
        self.completion_widget.setup_appearance(size, font)

    def set_indent_chars(self, indent_chars):
        """

        :param indent_chars:
        """
        self.indent_chars = indent_chars

    def set_tab_stop_width_spaces(self, tab_stop_width_spaces):
        """

        :param tab_stop_width_spaces:
        """
        self.tab_stop_width_spaces = tab_stop_width_spaces
        self.update_tab_stop_width_spaces()

    def update_tab_stop_width_spaces(self):
        """ """
        self.setTabStopWidth(self.fontMetrics().width(" " * self.tab_stop_width_spaces))

    def set_palette(self, background, foreground):
        """
        Set text editor palette colors:
        background color and caret (text cursor) color
        """
        palette = QPalette()
        # palette.setColor(QPalette.Base, background)
        palette.setColor(QPalette.Text, foreground)
        palette.setColor(QPalette.Window, background)
        self.setPalette(palette)

        # Set the right background color when changing color schemes
        # or creating new Editor windows. This seems to be a Qt bug.
        # Fixes Issue 2028
        if sys.platform == "darwin":
            if self.objectName():
                style = "QPlainTextEdit#%s {background: %s; color: %s;}" % (
                    self.objectName(),
                    background.name(),
                    foreground.name(),
                )
                self.setStyleSheet(style)

    # ------Extra selections
    def extra_selection_length(self, key):
        """

        :param key:
        :return:
        """
        selection = self.get_extra_selections(key)
        if selection:
            cursor = self.extra_selections_dict[key][0].cursor
            selection_length = cursor.selectionEnd() - cursor.selectionStart()
            return selection_length
        else:
            return 0

    def get_extra_selections(self, key):
        """

        :param key:
        :return:
        """
        return self.extra_selections_dict.get(key, [])

    def set_extra_selections(self, key, extra_selections):
        """

        :param key:
        :param extra_selections:
        """
        self.extra_selections_dict[key] = extra_selections
        self.extra_selections_dict = OrderedDict(
            sorted(
                self.extra_selections_dict.items(),
                key=lambda s: self.extra_selection_length(s[0]),
                reverse=True,
            )
        )

    def update_extra_selections(self):
        """ """
        extra_selections = []

        # Python 3 compatibility (weird): current line has to be
        # highlighted first
        if "current_cell" in self.extra_selections_dict:
            extra_selections.extend(self.extra_selections_dict["current_cell"])
        if "current_line" in self.extra_selections_dict:
            extra_selections.extend(self.extra_selections_dict["current_line"])

        for key, extra in list(self.extra_selections_dict.items()):
            if not (key == "current_line" or key == "current_cell"):
                extra_selections.extend(extra)
        self.setExtraSelections(extra_selections)

    def clear_extra_selections(self, key):
        """

        :param key:
        """
        self.extra_selections_dict[key] = []
        self.update_extra_selections()

    def changed(self):
        """Emit changed signal"""
        self.modificationChanged.emit(self.document().isModified())

    # ------Highlight current line
    def highlight_current_line(self):
        """Highlight current line"""
        selection = QTextEdit.ExtraSelection()
        selection.format.setProperty(QTextFormat.FullWidthSelection, True)
        selection.format.setBackground(self.currentline_color)
        selection.cursor = self.textCursor()
        selection.cursor.clearSelection()
        self.set_extra_selections("current_line", [selection])
        self.update_extra_selections()

    def unhighlight_current_line(self):
        """Unhighlight current line"""
        self.clear_extra_selections("current_line")

    # ------Highlight current cell
    def highlight_current_cell(self):
        """Highlight current cell"""
        if self.cell_separators is None or not self.highlight_current_cell_enabled:
            return
        selection = QTextEdit.ExtraSelection()
        selection.format.setProperty(QTextFormat.FullWidthSelection, True)
        selection.format.setBackground(self.currentcell_color)
        (
            selection.cursor,
            whole_file_selected,
            whole_screen_selected,
        ) = self.select_current_cell_in_visible_portion()
        if whole_file_selected:
            self.clear_extra_selections("current_cell")
        elif whole_screen_selected:
            if self.has_cell_separators:
                self.set_extra_selections("current_cell", [selection])
                self.update_extra_selections()
            else:
                self.clear_extra_selections("current_cell")
        else:
            self.set_extra_selections("current_cell", [selection])
            self.update_extra_selections()

    def unhighlight_current_cell(self):
        """Unhighlight current cell"""
        self.clear_extra_selections("current_cell")

    # ------Brace matching
    def find_brace_match(self, position, brace, forward):
        """

        :param position:
        :param brace:
        :param forward:
        :return:
        """
        start_pos, end_pos = self.BRACE_MATCHING_SCOPE
        if forward:
            bracemap = {"(": ")", "[": "]", "{": "}"}
            text = self.get_text(position, end_pos)
            i_start_open = 1
            i_start_close = 1
        else:
            bracemap = {")": "(", "]": "[", "}": "{"}
            text = self.get_text(start_pos, position)
            i_start_open = len(text) - 1
            i_start_close = len(text) - 1

        while True:
            if forward:
                i_close = text.find(bracemap[brace], i_start_close)
            else:
                i_close = text.rfind(bracemap[brace], 0, i_start_close + 1)
            if i_close > -1:
                if forward:
                    i_start_close = i_close + 1
                    i_open = text.find(brace, i_start_open, i_close)
                else:
                    i_start_close = i_close - 1
                    i_open = text.rfind(brace, i_close, i_start_open + 1)
                if i_open > -1:
                    if forward:
                        i_start_open = i_open + 1
                    else:
                        i_start_open = i_open - 1
                else:
                    # found matching brace
                    if forward:
                        return position + i_close
                    else:
                        return position - (len(text) - i_close)
            else:
                # no matching brace
                return

    def __highlight(self, positions, color=None, cancel=False):
        if cancel:
            self.clear_extra_selections("brace_matching")
            return
        extra_selections = []
        for position in positions:
            if position > self.get_position("eof"):
                return
            selection = QTextEdit.ExtraSelection()
            selection.format.setBackground(color)
            selection.cursor = self.textCursor()
            selection.cursor.clearSelection()
            selection.cursor.setPosition(position)
            selection.cursor.movePosition(
                QTextCursor.NextCharacter, QTextCursor.KeepAnchor
            )
            extra_selections.append(selection)
        self.set_extra_selections("brace_matching", extra_selections)
        self.update_extra_selections()

    def cursor_position_changed(self):
        """Brace matching"""
        if self.bracepos is not None:
            self.__highlight(self.bracepos, cancel=True)
            self.bracepos = None
        cursor = self.textCursor()
        if cursor.position() == 0:
            return
        cursor.movePosition(QTextCursor.PreviousCharacter, QTextCursor.KeepAnchor)
        text = str(cursor.selectedText())
        pos1 = cursor.position()
        if text in (")", "]", "}"):
            pos2 = self.find_brace_match(pos1, text, forward=False)
        elif text in ("(", "[", "{"):
            pos2 = self.find_brace_match(pos1, text, forward=True)
        else:
            return
        if pos2 is not None:
            self.bracepos = (pos1, pos2)
            self.__highlight(self.bracepos, color=self.matched_p_color)
        else:
            self.bracepos = (pos1,)
            self.__highlight(self.bracepos, color=self.unmatched_p_color)

    # -----Widget setup and options
    def set_codecompletion_auto(self, state):
        """Set code completion state"""
        self.codecompletion_auto = state

    def set_codecompletion_case(self, state):
        """Case sensitive completion"""
        self.codecompletion_case = state
        self.completion_widget.case_sensitive = state

    def set_codecompletion_enter(self, state):
        """Enable Enter key to select completion"""
        self.codecompletion_enter = state
        self.completion_widget.enter_select = state

    def set_calltips(self, state):
        """Set calltips state"""
        self.calltips = state

    def set_wrap_mode(self, mode=None):
        """
        Set wrap mode
        Valid *mode* values: None, 'word', 'character'
        """
        if mode == "word":
            wrap_mode = QTextOption.WrapAtWordBoundaryOrAnywhere
        elif mode == "character":
            wrap_mode = QTextOption.WrapAnywhere
        else:
            wrap_mode = QTextOption.NoWrap
        self.setWordWrapMode(wrap_mode)

    # ------Reimplementing Qt methods
    @Slot()
    def copy(self):
        """
        Reimplement Qt method
        Copy text to clipboard with correct EOL chars
        """
        if self.get_selected_text():
            QApplication.clipboard().setText(self.get_selected_text())

    def toPlainText(self):
        """
        Reimplement Qt method
        Fix PyQt4 bug on Windows and Python 3
        """
        # Fix what appears to be a PyQt4 bug when getting file
        # contents under Windows and PY3. This bug leads to
        # corruptions when saving files with certain combinations
        # of unicode chars on them (like the one attached on
        # Issue 1546)
        if os.name == "nt":
            text = self.get_text("sof", "eof")
            return (
                text.replace("\u2028", "\n")
                .replace("\u2029", "\n")
                .replace("\u0085", "\n")
            )
        else:
            return super().toPlainText()

    def keyPressEvent(self, event):
        """

        :param event:
        """
        _text, key = event.text(), event.key()
        ctrl = event.modifiers() & Qt.ControlModifier
        meta = event.modifiers() & Qt.MetaModifier
        # Use our own copy method for {Ctrl,Cmd}+C to avoid Qt
        # copying text in HTML (See Issue 2285)
        if (ctrl or meta) and key == Qt.Key_C:
            self.copy()
        else:
            super().keyPressEvent(event)

    # ------Text: get, set, ...
    def get_selection_as_executable_code(self):
        """Return selected text as a processed text,
        to be executable in a Python/IPython interpreter"""
        ls = self.get_line_separator()

        _indent = lambda line: len(line) - len(line.lstrip())

        line_from, line_to = self.get_selection_bounds()
        text = self.get_selected_text()
        if not text:
            return

        lines = text.split(ls)
        if len(lines) > 1:
            # Multiline selection -> eventually fixing indentation
            original_indent = _indent(self.get_text_line(line_from))
            text = (" " * (original_indent - _indent(lines[0]))) + text

        # If there is a common indent to all lines, find it.
        # Moving from bottom line to top line ensures that blank
        # lines inherit the indent of the line *below* it,
        # which is the desired behavior.
        min_indent = 999
        current_indent = 0
        lines = text.split(ls)
        for i in range(len(lines) - 1, -1, -1):
            line = lines[i]
            if line.strip():
                current_indent = _indent(line)
                min_indent = min(current_indent, min_indent)
            else:
                lines[i] = " " * current_indent
        if min_indent:
            lines = [line[min_indent:] for line in lines]

        # Remove any leading whitespace or comment lines
        # since they confuse the reserved word detector that follows below
        while lines:
            first_line = lines[0].lstrip()
            if first_line == "" or first_line[0] == "#":
                lines.pop(0)
            else:
                break

        # Add an EOL character after indentation blocks that start with some
        # Python reserved words, so that it gets evaluated automatically
        # by the console
        varname = re.compile(r"[a-zA-Z0-9_]*")  # Matches valid variable names.
        maybe = False
        nextexcept = ()
        for n, line in enumerate(lines):
            if not _indent(line):
                word = varname.match(line).group()
                if maybe and word not in nextexcept:
                    lines[n - 1] += ls
                    maybe = False
                if word:
                    if word in ("def", "for", "while", "with", "class"):
                        maybe = True
                        nextexcept = ()
                    elif word == "if":
                        maybe = True
                        nextexcept = ("elif", "else")
                    elif word == "try":
                        maybe = True
                        nextexcept = ("except", "finally")
        if maybe:
            if lines[-1].strip() == "":
                lines[-1] += ls
            else:
                lines.append(ls)

        return ls.join(lines)

    def __exec_cell(self):
        init_cursor = QTextCursor(self.textCursor())
        start_pos, end_pos = self.__save_selection()
        cursor, whole_file_selected = self.select_current_cell()
        if not whole_file_selected:
            self.setTextCursor(cursor)
        text = self.get_selection_as_executable_code()
        self.last_cursor_cell = init_cursor
        self.__restore_selection(start_pos, end_pos)
        if text is not None:
            text = text.rstrip()
        return text

    def get_cell_as_executable_code(self):
        """Return cell contents as executable code"""
        return self.__exec_cell()

    def get_last_cell_as_executable_code(self):
        """

        :return:
        """
        text = None
        if self.last_cursor_cell:
            self.setTextCursor(self.last_cursor_cell)
            self.highlight_current_cell()
            text = self.__exec_cell()
        return text

    def is_cell_separator(self, cursor=None, block=None):
        """Return True if cursor (or text block) is on a block separator"""
        assert cursor is not None or block is not None
        if cursor is not None:
            cursor0 = QTextCursor(cursor)
            cursor0.select(QTextCursor.BlockUnderCursor)
            text = str(cursor0.selectedText())
        else:
            text = str(block.text())
        if self.cell_separators is None:
            return False
        else:
            return text.lstrip().startswith(self.cell_separators)

    def select_current_cell(self):
        """Select cell under cursor
        cell = group of lines separated by CELL_SEPARATORS
        returns the textCursor and a boolean indicating if the
        entire file is selected"""
        cursor = self.textCursor()
        cursor.movePosition(QTextCursor.StartOfBlock)
        cur_pos = prev_pos = cursor.position()

        # Moving to the next line that is not a separator, if we are
        # exactly at one of them
        while self.is_cell_separator(cursor):
            cursor.movePosition(QTextCursor.NextBlock)
            prev_pos = cur_pos
            cur_pos = cursor.position()
            if cur_pos == prev_pos:
                return cursor, False
        prev_pos = cur_pos
        # If not, move backwards to find the previous separator
        while not self.is_cell_separator(cursor):
            cursor.movePosition(QTextCursor.PreviousBlock)
            prev_pos = cur_pos
            cur_pos = cursor.position()
            if cur_pos == prev_pos:
                if self.is_cell_separator(cursor):
                    return cursor, False
                else:
                    break
        cursor.setPosition(prev_pos)
        cell_at_file_start = cursor.atStart()
        # Once we find it (or reach the beginning of the file)
        # move to the next separator (or the end of the file)
        # so we can grab the cell contents
        while not self.is_cell_separator(cursor):
            cursor.movePosition(QTextCursor.NextBlock, QTextCursor.KeepAnchor)
            cur_pos = cursor.position()
            if cur_pos == prev_pos:
                cursor.movePosition(QTextCursor.EndOfBlock, QTextCursor.KeepAnchor)
                break
            prev_pos = cur_pos
        cell_at_file_end = cursor.atEnd()
        return cursor, cell_at_file_start and cell_at_file_end

    def select_current_cell_in_visible_portion(self):
        """Select cell under cursor in the visible portion of the file
        cell = group of lines separated by CELL_SEPARATORS

        Returns:
        - the textCursor
        - a boolean indicating if the entire file is selected
        - a boolean indicating if the entire visible portion of the file is selected"""
        cursor = self.textCursor()
        cursor.movePosition(QTextCursor.StartOfBlock)
        cur_pos = prev_pos = cursor.position()

        beg_pos = self.cursorForPosition(QPoint(0, 0)).position()
        bottom_right = QPoint(self.viewport().width() - 1, self.viewport().height() - 1)
        end_pos = self.cursorForPosition(bottom_right).position()

        # Moving to the next line that is not a separator, if we are
        # exactly at one of them
        while self.is_cell_separator(cursor):
            cursor.movePosition(QTextCursor.NextBlock)
            prev_pos = cur_pos
            cur_pos = cursor.position()
            if cur_pos == prev_pos:
                return cursor, False, False
        prev_pos = cur_pos
        # If not, move backwards to find the previous separator
        while not self.is_cell_separator(cursor) and cursor.position() >= beg_pos:
            cursor.movePosition(QTextCursor.PreviousBlock)
            prev_pos = cur_pos
            cur_pos = cursor.position()
            if cur_pos == prev_pos:
                if self.is_cell_separator(cursor):
                    return cursor, False, False
                else:
                    break
        cell_at_screen_start = cursor.position() <= beg_pos
        cursor.setPosition(prev_pos)
        cell_at_file_start = cursor.atStart()
        # Selecting cell header
        if not cell_at_file_start:
            cursor.movePosition(QTextCursor.PreviousBlock)
            cursor.movePosition(QTextCursor.NextBlock, QTextCursor.KeepAnchor)
        # Once we find it (or reach the beginning of the file)
        # move to the next separator (or the end of the file)
        # so we can grab the cell contents
        while not self.is_cell_separator(cursor) and cursor.position() <= end_pos:
            cursor.movePosition(QTextCursor.NextBlock, QTextCursor.KeepAnchor)
            cur_pos = cursor.position()
            if cur_pos == prev_pos:
                cursor.movePosition(QTextCursor.EndOfBlock, QTextCursor.KeepAnchor)
                break
            prev_pos = cur_pos
        cell_at_file_end = cursor.atEnd()
        cell_at_screen_end = cursor.position() >= end_pos
        return (
            cursor,
            cell_at_file_start and cell_at_file_end,
            cell_at_screen_start and cell_at_screen_end,
        )

    def go_to_next_cell(self):
        """Go to the next cell of lines"""
        cursor = self.textCursor()
        cursor.movePosition(QTextCursor.NextBlock)
        cur_pos = prev_pos = cursor.position()
        while not self.is_cell_separator(cursor):
            # Moving to the next code cell
            cursor.movePosition(QTextCursor.NextBlock)
            prev_pos = cur_pos
            cur_pos = cursor.position()
            if cur_pos == prev_pos:
                return
        self.setTextCursor(cursor)

    def go_to_previous_cell(self):
        """Go to the previous cell of lines"""
        cursor = self.textCursor()
        cur_pos = prev_pos = cursor.position()

        if self.is_cell_separator(cursor):
            # Move to the previous cell
            cursor.movePosition(QTextCursor.PreviousBlock)
            cur_pos = prev_pos = cursor.position()

        while not self.is_cell_separator(cursor):
            # Move to the previous cell or the beginning of the current cell
            cursor.movePosition(QTextCursor.PreviousBlock)
            prev_pos = cur_pos
            cur_pos = cursor.position()
            if cur_pos == prev_pos:
                return

        self.setTextCursor(cursor)

    def get_line_count(self):
        """Return document total line number"""
        return self.blockCount()

    def __save_selection(self):
        """Save current cursor selection and return position bounds"""
        cursor = self.textCursor()
        return cursor.selectionStart(), cursor.selectionEnd()

    def __restore_selection(self, start_pos, end_pos):
        """Restore cursor selection from position bounds"""
        cursor = self.textCursor()
        cursor.setPosition(start_pos)
        cursor.setPosition(end_pos, QTextCursor.KeepAnchor)
        self.setTextCursor(cursor)

    def __duplicate_line_or_selection(self, after_current_line=True):
        """Duplicate current line or selected text"""
        cursor = self.textCursor()
        cursor.beginEditBlock()
        start_pos, end_pos = self.__save_selection()
        if str(cursor.selectedText()):
            cursor.setPosition(end_pos)
            # Check if end_pos is at the start of a block: if so, starting
            # changes from the previous block
            cursor.movePosition(QTextCursor.StartOfBlock, QTextCursor.KeepAnchor)
            if not str(cursor.selectedText()):
                cursor.movePosition(QTextCursor.PreviousBlock)
                end_pos = cursor.position()

        cursor.setPosition(start_pos)
        cursor.movePosition(QTextCursor.StartOfBlock)
        while cursor.position() <= end_pos:
            cursor.movePosition(QTextCursor.EndOfBlock, QTextCursor.KeepAnchor)
            if cursor.atEnd():
                cursor_temp = QTextCursor(cursor)
                cursor_temp.clearSelection()
                cursor_temp.insertText(self.get_line_separator())
                break
            cursor.movePosition(QTextCursor.NextBlock, QTextCursor.KeepAnchor)
        text = cursor.selectedText()
        cursor.clearSelection()

        if not after_current_line:
            # Moving cursor before current line/selected text
            cursor.setPosition(start_pos)
            cursor.movePosition(QTextCursor.StartOfBlock)
            start_pos += len(text)
            end_pos += len(text)

        cursor.insertText(text)
        cursor.endEditBlock()
        self.setTextCursor(cursor)
        self.__restore_selection(start_pos, end_pos)

    def duplicate_line(self):
        """
        Duplicate current line or selected text
        Paste the duplicated text *after* the current line/selected text
        """
        self.__duplicate_line_or_selection(after_current_line=True)

    def copy_line(self):
        """
        Copy current line or selected text
        Paste the duplicated text *before* the current line/selected text
        """
        self.__duplicate_line_or_selection(after_current_line=False)

    def __move_line_or_selection(self, after_current_line=True):
        """Move current line or selected text"""
        cursor = self.textCursor()
        cursor.beginEditBlock()
        start_pos, end_pos = self.__save_selection()
        last_line = False

        # ------ Select text

        # Get selection start location
        cursor.setPosition(start_pos)
        cursor.movePosition(QTextCursor.StartOfBlock)
        start_pos = cursor.position()

        # Get selection end location
        cursor.setPosition(end_pos)
        if not cursor.atBlockStart() or end_pos == start_pos:
            cursor.movePosition(QTextCursor.EndOfBlock)
            cursor.movePosition(QTextCursor.NextBlock)
        end_pos = cursor.position()

        # Check if selection ends on the last line of the document
        if cursor.atEnd():
            if not cursor.atBlockStart() or end_pos == start_pos:
                last_line = True

        # ------ Stop if at document boundary

        cursor.setPosition(start_pos)
        if cursor.atStart() and not after_current_line:
            # Stop if selection is already at top of the file while moving up
            cursor.endEditBlock()
            self.setTextCursor(cursor)
            self.__restore_selection(start_pos, end_pos)
            return

        cursor.setPosition(end_pos, QTextCursor.KeepAnchor)
        if last_line and after_current_line:
            # Stop if selection is already at end of the file while moving down
            cursor.endEditBlock()
            self.setTextCursor(cursor)
            self.__restore_selection(start_pos, end_pos)
            return

        # ------ Move text

        sel_text = str(cursor.selectedText())
        cursor.removeSelectedText()

        if after_current_line:
            # Shift selection down
            text = str(cursor.block().text())
            sel_text = os.linesep + sel_text[0:-1]  # Move linesep at the start
            cursor.movePosition(QTextCursor.EndOfBlock)
            start_pos += len(text) + 1
            end_pos += len(text)
            if not cursor.atEnd():
                end_pos += 1
        else:
            # Shift selection up
            if last_line:
                # Remove the last linesep and add it to the selected text
                cursor.deletePreviousChar()
                sel_text = sel_text + os.linesep
                cursor.movePosition(QTextCursor.StartOfBlock)
                end_pos += 1
            else:
                cursor.movePosition(QTextCursor.PreviousBlock)
            text = str(cursor.block().text())
            start_pos -= len(text) + 1
            end_pos -= len(text) + 1

        cursor.insertText(sel_text)

        cursor.endEditBlock()
        self.setTextCursor(cursor)
        self.__restore_selection(start_pos, end_pos)

    def move_line_up(self):
        """Move up current line or selected text"""
        self.__move_line_or_selection(after_current_line=False)

    def move_line_down(self):
        """Move down current line or selected text"""
        self.__move_line_or_selection(after_current_line=True)

    def go_to_new_line(self):
        """Go to the end of the current line and create a new line"""
        self.stdkey_end(False, False)
        self.insert_text(self.get_line_separator())

    def extend_selection_to_complete_lines(self):
        """Extend current selection to complete lines"""
        cursor = self.textCursor()
        start_pos, end_pos = cursor.selectionStart(), cursor.selectionEnd()
        cursor.setPosition(start_pos)
        cursor.setPosition(end_pos, QTextCursor.KeepAnchor)
        if cursor.atBlockStart():
            cursor.movePosition(QTextCursor.PreviousBlock, QTextCursor.KeepAnchor)
            cursor.movePosition(QTextCursor.EndOfBlock, QTextCursor.KeepAnchor)
        self.setTextCursor(cursor)

    def delete_line(self):
        """Delete current line"""
        cursor = self.textCursor()
        if self.has_selected_text():
            self.extend_selection_to_complete_lines()
            start_pos, end_pos = cursor.selectionStart(), cursor.selectionEnd()
            cursor.setPosition(start_pos)
        else:
            start_pos = end_pos = cursor.position()
        cursor.beginEditBlock()
        cursor.setPosition(start_pos)
        cursor.movePosition(QTextCursor.StartOfBlock)
        while cursor.position() <= end_pos:
            cursor.movePosition(QTextCursor.EndOfBlock, QTextCursor.KeepAnchor)
            if cursor.atEnd():
                break
            cursor.movePosition(QTextCursor.NextBlock, QTextCursor.KeepAnchor)
        cursor.removeSelectedText()
        cursor.endEditBlock()
        self.ensureCursorVisible()

    def set_selection(self, start, end):
        """

        :param start:
        :param end:
        """
        cursor = self.textCursor()
        cursor.setPosition(start)
        cursor.setPosition(end, QTextCursor.KeepAnchor)
        self.setTextCursor(cursor)

    def truncate_selection(self, position_from):
        """Unselect read-only parts in shell, like prompt"""
        position_from = self.get_position(position_from)
        cursor = self.textCursor()
        start, end = cursor.selectionStart(), cursor.selectionEnd()
        if start < end:
            start = max([position_from, start])
        else:
            end = max([position_from, end])
        self.set_selection(start, end)

    def restrict_cursor_position(self, position_from, position_to):
        """In shell, avoid editing text except between prompt and EOF"""
        position_from = self.get_position(position_from)
        position_to = self.get_position(position_to)
        cursor = self.textCursor()
        cursor_position = cursor.position()
        if cursor_position < position_from or cursor_position > position_to:
            self.set_cursor_position(position_to)

    # ------Code completion / Calltips
    def hide_tooltip_if_necessary(self, key):
        """Hide calltip when necessary"""
        try:
            calltip_char = self.get_character(self.calltip_position)
            before = self.is_cursor_before(self.calltip_position, char_offset=1)
            other = key in (Qt.Key_ParenRight, Qt.Key_Period, Qt.Key_Tab)
            if calltip_char not in ("?", "(") or before or other:
                QToolTip.hideText()
        except (IndexError, TypeError):
            QToolTip.hideText()

    def show_completion_widget(self, textlist, automatic=True):
        """Show completion widget"""
        self.completion_widget.show_list(textlist, automatic=automatic)

    def hide_completion_widget(self):
        """Hide completion widget"""
        self.completion_widget.hide()

    def show_completion_list(self, completions, completion_text="", automatic=True):
        """Display the possible completions"""
        if not completions:
            return
        if not isinstance(completions[0], tuple):
            completions = [(c, "") for c in completions]
        if len(completions) == 1 and completions[0][0] == completion_text:
            return
        self.completion_text = completion_text
        # Sorting completion list (entries starting with underscore are
        # put at the end of the list):
        underscore = set(
            [(comp, t) for (comp, t) in completions if comp.startswith("_")]
        )

        completions = sorted(set(completions) - underscore, key=lambda x: x[0].lower())
        completions += sorted(underscore, key=lambda x: x[0].lower())
        self.show_completion_widget(completions, automatic=automatic)

    def select_completion_list(self):
        """Completion list is active, Enter was just pressed"""
        self.completion_widget.item_selected()

    def insert_completion(self, text):
        """

        :param text:
        """
        if text:
            cursor = self.textCursor()
            cursor.movePosition(
                QTextCursor.PreviousCharacter,
                QTextCursor.KeepAnchor,
                len(self.completion_text),
            )
            cursor.removeSelectedText()
            self.insert_text(text)

    def is_completion_widget_visible(self):
        """Return True is completion list widget is visible"""
        return self.completion_widget.isVisible()

    # ------Standard keys
    def stdkey_clear(self):
        """ """
        if not self.has_selected_text():
            self.moveCursor(QTextCursor.NextCharacter, QTextCursor.KeepAnchor)
        self.remove_selected_text()

    def stdkey_backspace(self):
        """ """
        if not self.has_selected_text():
            self.moveCursor(QTextCursor.PreviousCharacter, QTextCursor.KeepAnchor)
        self.remove_selected_text()

    def __get_move_mode(self, shift):
        return QTextCursor.KeepAnchor if shift else QTextCursor.MoveAnchor

    def stdkey_up(self, shift):
        """

        :param shift:
        """
        self.moveCursor(QTextCursor.Up, self.__get_move_mode(shift))

    def stdkey_down(self, shift):
        """

        :param shift:
        """
        self.moveCursor(QTextCursor.Down, self.__get_move_mode(shift))

    def stdkey_tab(self):
        """ """
        self.insert_text(self.indent_chars)

    def stdkey_home(self, shift, ctrl, prompt_pos=None):
        """Smart HOME feature: cursor is first moved at
        indentation position, then at the start of the line"""
        move_mode = self.__get_move_mode(shift)
        if ctrl:
            self.moveCursor(QTextCursor.Start, move_mode)
        else:
            cursor = self.textCursor()
            if prompt_pos is None:
                start_position = self.get_position("sol")
            else:
                start_position = self.get_position(prompt_pos)
            text = self.get_text(start_position, "eol")
            indent_pos = start_position + len(text) - len(text.lstrip())
            if cursor.position() != indent_pos:
                cursor.setPosition(indent_pos, move_mode)
            else:
                cursor.setPosition(start_position, move_mode)
            self.setTextCursor(cursor)

    def stdkey_end(self, shift, ctrl):
        """

        :param shift:
        :param ctrl:
        """
        move_mode = self.__get_move_mode(shift)
        if ctrl:
            self.moveCursor(QTextCursor.End, move_mode)
        else:
            self.moveCursor(QTextCursor.EndOfBlock, move_mode)

    def stdkey_pageup(self):
        """ """
        pass

    def stdkey_pagedown(self):
        """ """
        pass

    def stdkey_escape(self):
        """ """
        pass

    # ----Qt Events
    def mousePressEvent(self, event):
        """Reimplement Qt method"""
        if sys.platform.startswith("linux") and event.button() == Qt.MidButton:
            self.calltip_widget.hide()
            self.setFocus()
            event = QMouseEvent(
                QEvent.MouseButtonPress,
                event.pos(),
                Qt.LeftButton,
                Qt.LeftButton,
                Qt.NoModifier,
            )
            QPlainTextEdit.mousePressEvent(self, event)
            QPlainTextEdit.mouseReleaseEvent(self, event)
            # Send selection text to clipboard to be able to use
            # the paste method and avoid the strange Issue 1445
            # NOTE: This issue seems a focusing problem but it
            # seems really hard to track
            mode_clip = QClipboard.Clipboard
            mode_sel = QClipboard.Selection
            text_clip = QApplication.clipboard().text(mode=mode_clip)
            text_sel = QApplication.clipboard().text(mode=mode_sel)
            QApplication.clipboard().setText(text_sel, mode=mode_clip)
            self.paste()
            QApplication.clipboard().setText(text_clip, mode=mode_clip)
        else:
            self.calltip_widget.hide()
            QPlainTextEdit.mousePressEvent(self, event)

    def focusInEvent(self, event):
        """Reimplemented to handle focus"""
        self.focus_changed.emit()
        self.focus_in.emit()
        self.highlight_current_cell()
        QPlainTextEdit.focusInEvent(self, event)

    def focusOutEvent(self, event):
        """Reimplemented to handle focus"""
        self.focus_changed.emit()
        QPlainTextEdit.focusOutEvent(self, event)

    def wheelEvent(self, event):
        """Reimplemented to emit zoom in/out signals when Ctrl is pressed"""
        # This feature is disabled on MacOS, see Issue 1510
        if sys.platform != "darwin":
            if event.modifiers() & Qt.ControlModifier:
                if hasattr(event, "angleDelta"):
                    if event.angleDelta().y() < 0:
                        self.zoom_out.emit()
                    elif event.angleDelta().y() > 0:
                        self.zoom_in.emit()
                elif hasattr(event, "delta"):
                    if event.delta() < 0:
                        self.zoom_out.emit()
                    elif event.delta() > 0:
                        self.zoom_in.emit()
                return
        QPlainTextEdit.wheelEvent(self, event)
        self.highlight_current_cell()


class QtANSIEscapeCodeHandler(ANSIEscapeCodeHandler):
    """ """

    def __init__(self):
        ANSIEscapeCodeHandler.__init__(self)
        self.base_format = None
        self.current_format = None

    def set_light_background(self, state):
        """

        :param state:
        """
        if state:
            self.default_foreground_color = 30
            self.default_background_color = 47
        else:
            self.default_foreground_color = 37
            self.default_background_color = 40

    def set_base_format(self, base_format):
        """

        :param base_format:
        """
        self.base_format = base_format

    def get_format(self):
        """

        :return:
        """
        return self.current_format

    def set_style(self):
        """
        Set font style with the following attributes:
        'foreground_color', 'background_color', 'italic',
        'bold' and 'underline'
        """
        if self.current_format is None:
            assert self.base_format is not None
            self.current_format = QTextCharFormat(self.base_format)
        # Foreground color
        if self.foreground_color is None:
            qcolor = self.base_format.foreground()
        else:
            cstr = self.ANSI_COLORS[self.foreground_color - 30][self.intensity]
            qcolor = QColor(cstr)
        self.current_format.setForeground(qcolor)
        # Background color
        if self.background_color is None:
            qcolor = self.base_format.background()
        else:
            cstr = self.ANSI_COLORS[self.background_color - 40][self.intensity]
            qcolor = QColor(cstr)
        self.current_format.setBackground(qcolor)

        font = self.current_format.font()
        # Italic
        if self.italic is None:
            italic = self.base_format.fontItalic()
        else:
            italic = self.italic
        font.setItalic(italic)
        # Bold
        if self.bold is None:
            bold = self.base_format.font().bold()
        else:
            bold = self.bold
        font.setBold(bold)
        # Underline
        if self.underline is None:
            underline = self.base_format.font().underline()
        else:
            underline = self.underline
        font.setUnderline(underline)
        self.current_format.setFont(font)


def inverse_color(color):
    """Inverse color"""
    inv_color = QColor()
    inv_color.setHsv(color.hue(), color.saturation(), 255 - color.value())
    return inv_color


class ConsoleFontStyle(object):
    """Console font style management"""

    def __init__(self, foregroundcolor, backgroundcolor, bold, italic, underline):
        self.foregroundcolor = foregroundcolor
        self.backgroundcolor = backgroundcolor
        self.bold = bold
        self.italic = italic
        self.underline = underline
        self.format = None

    def apply_style(self, font, is_default):
        """Apply style to font

        Args:
            font: QFont
            is_default: Default is standard text (not error, link, etc.)
        """
        self.format = QTextCharFormat()
        self.format.setFont(font)
        fg_color = QColor(self.foregroundcolor)
        if is_default:
            self.format.setForeground(
                inverse_color(fg_color) if qth.is_dark_theme() else fg_color
            )
        else:
            self.format.setForeground(fg_color)
        self.format.setBackground(qth.get_background_color())
        font = self.format.font()
        font.setBold(self.bold)
        font.setItalic(self.italic)
        font.setUnderline(self.underline)
        self.format.setFont(font)


class ConsoleBaseWidget(TextEditBaseWidget):
    """Console base widget"""

    BRACE_MATCHING_SCOPE = ("sol", "eol")
    COLOR_PATTERN = re.compile(r"\x01?\x1b\[(.*?)m\x02?")
    exception_occurred = Signal(str, bool)
    userListActivated = Signal(int, str)
    completion_widget_activated = Signal(str)

    def __init__(self, parent=None):
        TextEditBaseWidget.__init__(self, parent)
        self.setMaximumBlockCount(300)

        # ANSI escape code handler
        self.ansi_handler = QtANSIEscapeCodeHandler()

        # Disable undo/redo (nonsense for a console widget...):
        self.setUndoRedoEnabled(False)

        self.userListActivated.connect(
            lambda user_id, text: self.completion_widget_activated.emit(text)
        )

        self.default_style = ConsoleFontStyle(
            foregroundcolor=0x000000,
            backgroundcolor=0xFFFFFF,
            bold=False,
            italic=False,
            underline=False,
        )
        self.error_style = ConsoleFontStyle(
            foregroundcolor=0xFF0000,
            backgroundcolor=0xFFFFFF,
            bold=False,
            italic=False,
            underline=False,
        )
        self.traceback_link_style = ConsoleFontStyle(
            foregroundcolor=0x0000FF,
            backgroundcolor=0xFFFFFF,
            bold=True,
            italic=False,
            underline=True,
        )
        self.prompt_style = ConsoleFontStyle(
            foregroundcolor=0x00AA00,
            backgroundcolor=0xFFFFFF,
            bold=True,
            italic=False,
            underline=False,
        )
        self.font_styles = (
            self.default_style,
            self.error_style,
            self.traceback_link_style,
            self.prompt_style,
        )
        self.update_color_mode()
        self.setMouseTracking(True)

    def update_color_mode(self):
        """Update color mode according to the current theme"""
        self.set_light_background(not qth.is_dark_theme())

    def set_light_background(self, state):
        """Set light background state

        Args:
            state: bool
        """
        default_fg_color = QColor(self.default_style.foregroundcolor)
        if not state:
            default_fg_color = inverse_color(default_fg_color)
        bg_color = qth.get_background_color()

        cursor = self.textCursor()
        cursor.movePosition(QTextCursor.Start)
        while not cursor.atEnd():
            cursor.setPosition(cursor.block().position())
            cursor.movePosition(QTextCursor.EndOfBlock, QTextCursor.KeepAnchor)
            charformat = cursor.charFormat()
            if charformat.background() == bg_color:
                break
            charformat.setBackground(bg_color)
            col = charformat.foreground().color()
            if col.red() == col.green() and col.green() == col.blue():
                # Default style
                charformat.setForeground(default_fg_color)
            cursor.setCharFormat(charformat)
            cursor.movePosition(QTextCursor.NextBlock)

        self.set_palette(background=bg_color, foreground=qth.get_foreground_color())
        self.ansi_handler.set_light_background(state)
        self.set_pythonshell_font()

    # ------Python shell
    def insert_text(self, text):
        """Reimplement TextEditBaseWidget method"""
        # Eventually this maybe should wrap to insert_text_to if
        # backspace-handling is required
        self.textCursor().insertText(text, self.default_style.format)

    def paste(self):
        """Reimplement Qt method"""
        if self.has_selected_text():
            self.remove_selected_text()
        self.insert_text(QApplication.clipboard().text())

    def append_text_to_shell(self, text, error, prompt):
        """
        Append text to Python shell
        In a way, this method overrides the method 'insert_text' when text is
        inserted at the end of the text widget for a Python shell

        Handles error messages and show blue underlined links
        Handles ANSI color sequences
        Handles ANSI FF sequence
        """
        cursor = self.textCursor()
        cursor.movePosition(QTextCursor.End)
        if "\r" in text:  # replace \r\n with \n
            text = text.replace("\r\n", "\n")
            text = text.replace("\r", "\n")
        while True:
            index = text.find(chr(12))
            if index == -1:
                break
            text = text[index + 1 :]
            self.clear()
        if error:
            is_traceback = False
            for text in text.splitlines(True):
                if text.startswith("  File") and not text.startswith('  File "<'):
                    is_traceback = True
                    # Show error links in blue underlined text
                    cursor.insertText("  ", self.default_style.format)
                    cursor.insertText(text[2:], self.traceback_link_style.format)
                else:
                    # Show error/warning messages in red
                    cursor.insertText(text, self.error_style.format)
            self.exception_occurred.emit(text, is_traceback)
        elif prompt:
            # Show prompt in green
            insert_text_to(cursor, text, self.prompt_style.format)
        else:
            # Show other outputs in black
            last_end = 0
            for match in self.COLOR_PATTERN.finditer(text):
                insert_text_to(
                    cursor, text[last_end : match.start()], self.default_style.format
                )
                last_end = match.end()
                try:
                    for code in [int(_c) for _c in match.group(1).split(";")]:
                        self.ansi_handler.set_code(code)
                except ValueError:
                    pass
                self.default_style.format = self.ansi_handler.get_format()
            insert_text_to(cursor, text[last_end:], self.default_style.format)
        self.set_cursor_position("eof")
        self.setCurrentCharFormat(self.default_style.format)

    def set_pythonshell_font(self, font=None):
        """Python Shell only"""
        if font is None:
            if self.default_style.format is None:
                font = self.font()
            else:
                font = self.default_style.format.font()
        for style in self.font_styles:
            style.apply_style(font=font, is_default=style is self.default_style)
        self.ansi_handler.set_base_format(self.default_style.format)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""Validation modes tests"""

import os.path as osp

import numpy as np
import pytest

import guidata.dataset as gds
from guidata.config import ValidationMode, get_validation_mode, set_validation_mode
from guidata.env import execenv


def check_array(value: np.ndarray, raise_exception: bool = False) -> bool:
    """Check if value is a valid 2D array of floats.

    Args:
        value: value to check
        raise_exception: if True, raise an exception on invalid value

    Returns:
        True if value is valid, False otherwise
    """
    if (
        not isinstance(value, np.ndarray)
        or value.ndim != 2
        or not np.issubdtype(value.dtype, np.floating)
    ):
        if raise_exception:
            raise TypeError("Float array must be a 2D numpy array of floats")
        return False
    return True


class Parameters(gds.DataSet):
    """Example dataset"""

    fitem = gds.FloatItem("Float", min=1, max=250)
    iitem = gds.IntItem("Integer", max=20, nonzero=True)
    sitem = gds.StringItem("String", notempty=True)
    aitem = gds.FloatArrayItem("Array", check_callback=check_array)
    fileopenitem = gds.FileOpenItem("File", ("py",))
    filesopenitem = gds.FilesOpenItem("Files", ("py",))
    filesaveitem = gds.FileSaveItem("Save file", ("py",))
    directoryitem = gds.DirectoryItem("Directory")


VALID_DATA = {
    "fitem": 100.0,
    "iitem": 10,
    "sitem": "test",
    "aitem": np.array([[1.0, 2.0], [3.0, 4.0]]),
    "fileopenitem": __file__,
    "filesopenitem": [
        __file__,
        osp.join(osp.dirname(__file__), "__init__.py"),
    ],
    "filesaveitem": "test.py",
    "directoryitem": osp.dirname(__file__),
}

INVALID_DATA = {
    "fitem": (
        300.0,  # Out of range
        "10",  # Not a float
    ),
    "iitem": (
        30,  # Out of range
        0,  # Zero not allowed
        "test",  # Not an integer
        23.2323,  # Not an integer
    ),
    "aitem": (
        np.array([1.0, 2.0]),  # Not a 2D array
        np.array([[1, 2], [3, 4]]),  # Not a float array
    ),
    "sitem": (
        "",  # Empty string not allowed
        123,  # Not a string
    ),
    "fileopenitem": (
        "nonexistent.py",  # Nonexistent file
    ),
    "filesopenitem": (["nonexistent1.py", "nonexistent2.py"],),
    "filesaveitem": (
        "",  # Invalid empty file name
    ),
    "directoryitem": (
        "nonexistent_dir",  # Nonexistent directory
    ),
}


def test_default_validation_mode():
    """Test default validation mode"""
    execenv.print("Testing default validation mode: ", end="")
    assert get_validation_mode() == ValidationMode.DISABLED
    execenv.print("OK")


def __check_assigned_value_is_equal(assigned_value, expected_value):
    """Check if the assigned value is correctly set"""
    if isinstance(expected_value, np.ndarray):
        # For arrays, we check if the value is set correctly
        assert isinstance(assigned_value, np.ndarray)
        assert assigned_value.shape == expected_value.shape
        assert np.all(assigned_value == expected_value)
    else:
        # For other types, we check if the value is set correctly
        assert assigned_value == expected_value


def __check_assigned_value_is_not_equal(assigned_value, expected_value):
    """Check if the assigned value is not equal to the real value"""
    if isinstance(expected_value, np.ndarray):
        # For arrays, we check if the value is set correctly
        if isinstance(assigned_value, np.ndarray):
            assert assigned_value.shape == expected_value.shape
            assert not np.all(assigned_value == expected_value)
        else:
            assert assigned_value is None
    else:
        # For other types, we check if the value is set correctly
        assert assigned_value != expected_value


def test_valid_data():
    """Test valid data"""
    params = Parameters()
    execenv.print("Testing valid data: ", end="")
    for name, value in VALID_DATA.items():
        setattr(params, name, value)
        __check_assigned_value_is_equal(getattr(params, name), value)
    execenv.print("OK")


def test_invalid_data_with_no_validation():
    """Test invalid data with validation disabled"""
    old_mode = get_validation_mode()
    params = Parameters()
    set_validation_mode(ValidationMode.DISABLED)
    assert get_validation_mode() == ValidationMode.DISABLED
    execenv.print("Testing invalid data with validation disabled")
    for name, values in INVALID_DATA.items():
        for value in values:
            execenv.print(f"  Testing {name} with value: {value}")
            setattr(params, name, value)
            # No exception should be raised
            __check_assigned_value_is_equal(getattr(params, name), value)
    set_validation_mode(old_mode)


def test_invalid_data_with_enabled_validation():
    """Test invalid data with validation enabled"""
    old_mode = get_validation_mode()
    params = Parameters()
    set_validation_mode(ValidationMode.ENABLED)
    assert get_validation_mode() == ValidationMode.ENABLED
    execenv.print("Testing invalid data with validation enabled:")
    for name, values in INVALID_DATA.items():
        for value in values:
            execenv.print(f"  Testing {name} with value: {value}")
            # Check that a warning is raised
            with pytest.warns(gds.DataItemValidationWarning):
                setattr(params, name, value)
            # The value should be set anyway
            __check_assigned_value_is_equal(getattr(params, name), value)
    set_validation_mode(old_mode)


def test_invalid_data_with_strict_validation():
    """Test invalid data with strict validation"""
    old_mode = get_validation_mode()
    params = Parameters()
    set_validation_mode(ValidationMode.STRICT)
    assert get_validation_mode() == ValidationMode.STRICT
    execenv.print("Testing invalid data with strict validation:")
    for name, values in INVALID_DATA.items():
        for value in values:
            execenv.print(f"  Testing {name} with value: {value}")
            # Check that an exception is raised
            with pytest.raises(gds.DataItemValidationError):
                setattr(params, name, value)
            # The value should not be set
            __check_assigned_value_is_not_equal(getattr(params, name), value)
    set_validation_mode(old_mode)


if __name__ == "__main__":
    test_default_validation_mode()
    test_valid_data()
    test_invalid_data_with_no_validation()
    test_invalid_data_with_enabled_validation()
    test_invalid_data_with_strict_validation()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
rotatedlabel
------------

The ``guidata.widgets.rotatedlabel`` module provides a widget for displaying
rotated text.
"""

from math import cos, pi, sin

from qtpy.QtCore import QSize, Qt
from qtpy.QtGui import QFont, QPainter, QPen
from qtpy.QtWidgets import QLabel

from guidata.configtools import get_family


class RotatedLabel(QLabel):
    """
    Rotated QLabel
    (rich text is not supported)
    Arguments:
        * parent: parent widget
        * angle=270 (int): rotation angle in degrees
        * family (string): font family
        * bold (bool): font weight
        * italic (bool): font italic style
        * color (QColor): font color
    """

    def __init__(
        self,
        text,
        parent=None,
        angle=270,
        family=None,
        bold=False,
        italic=False,
        color=None,
    ):
        QLabel.__init__(self, text, parent)
        font = self.font()
        if family is not None:
            font.setFamily(get_family(family))
        font.setBold(bold)
        font.setItalic(italic)
        font.setHintingPreference(QFont.PreferNoHinting)
        self.setFont(font)
        self.color = color
        self.angle = angle
        self.setAlignment(Qt.AlignCenter)

    def paintEvent(self, evt):
        painter = QPainter(self)
        if self.color is not None:
            painter.setPen(QPen(self.color))
        painter.rotate(self.angle)
        transform = painter.transform().inverted()[0]
        rct = transform.mapRect(self.rect())
        painter.drawText(rct, self.alignment(), self.text())

    def sizeHint(self):
        hint = QLabel.sizeHint(self)
        width, height = hint.width(), hint.height()
        angle = self.angle * pi / 180
        rotated_width = int(abs(width * cos(angle)) + abs(height * sin(angle)))
        rotated_height = int(abs(width * sin(angle)) + abs(height * cos(angle)))
        return QSize(rotated_width, rotated_height)

    def minimumSizeHint(self):
        return self.sizeHint()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
All guidata DataItem objects demo

A DataSet object is a set of parameters of various types (integer, float,
boolean, string, etc.) which may be edited in a dialog box thanks to the
'edit' method. Parameters are defined by assigning DataItem objects to a
DataSet class definition: each parameter type has its own DataItem class
(IntItem for integers, FloatItem for floats, StringItem for strings, etc.)
"""

# guitest: show

import atexit
import datetime
import shutil
import tempfile
from enum import Enum

import numpy as np

import guidata.dataset as gds
from guidata.config import _
from guidata.env import execenv
from guidata.qthelpers import qt_app_context

# Creating temporary files and registering cleanup functions
TEMPDIR = tempfile.mkdtemp(prefix="test_")
atexit.register(shutil.rmtree, TEMPDIR)
FILE_ETA = tempfile.NamedTemporaryFile(suffix=".eta", dir=TEMPDIR)
atexit.register(FILE_ETA.close)
FILE_CSV = tempfile.NamedTemporaryFile(suffix=".csv", dir=TEMPDIR)
atexit.register(FILE_CSV.close)


class ImageTypes(Enum):
    """Image types."""

    #: Image filled with zeros
    ZEROS = _("Zeros")
    #: Empty image (filled with data from memory state)
    EMPTY = _("Empty")
    #: Image filled with random data (uniform law)
    UNIFORMRANDOM = _("Random (uniform law)")
    #: Image filled with random data (normal law)
    NORMALRANDOM = _("Random (normal law)")
    #: 2D Gaussian image
    GAUSS = _("Gaussian")
    #: Bilinear form image
    RAMP = _("2D ramp")


class FilteringMethod(gds.LabeledEnum):
    """LabeledEnum example with invariant keys and UI labels."""

    GAUSSIAN = "gaussian", _("Gaussian filter")
    MEDIAN = "median", _("Median filter")
    SOBEL = "sobel", _("Sobel edge detection")
    LAPLACIAN = "laplacian", _("Laplacian filter")
    BILATERAL = "bilateral", _("Bilateral filter")
    MORPHOLOGICAL = "morphological"  # Uses key as label


class ThresholdMode(gds.LabeledEnum):
    """Another LabeledEnum example."""

    OTSU = "otsu_method", _("Otsu's method")
    ADAPTIVE = "adaptive_mean", _("Adaptive mean")
    GLOBAL = "global_threshold", _("Global threshold")
    TRIANGLE = "triangle_algorithm", _("Triangle algorithm")


class Parameters(gds.DataSet):
    """
    DataSet test
    The following text is the DataSet 'comment': <br>Plain text or
    <b>rich text<sup>2</sup></b> are both supported,
    as well as special characters (α, β, γ, δ, ...)
    """

    dir = gds.DirectoryItem("Directory", default=TEMPDIR)
    preview = gds.TextItem("File names preview", default="-")
    option = gds.ChoiceItem("Option", (("1", "first choice"), ("2", "second choice")))

    fname = gds.FileOpenItem("Open file", ("csv", "eta"), FILE_CSV.name)
    fnames = gds.FilesOpenItem("Open files", "csv", [FILE_CSV.name])
    fname_s = gds.FileSaveItem("Save file", "eta", FILE_ETA.name)
    string = gds.StringItem("String", default="default string !?")
    text = gds.TextItem("Text", default="default\nmultiline\ntext")
    float_slider = gds.FloatItem(
        "Float (with slider)", default=0.5, min=0, max=1, step=0.01, slider=True
    )
    integer = gds.IntItem("Integer", default=5, min=3, max=16, slider=True).set_pos(
        col=1
    )
    dtime = gds.DateTimeItem("Date/time", default=datetime.datetime(2010, 10, 10))
    date = gds.DateItem("Date", default=datetime.date(2010, 10, 10)).set_pos(col=1)
    bool1 = gds.BoolItem("Boolean option without label")
    bool2 = gds.BoolItem("Boolean option with label", "Label")
    _bg = gds.BeginGroup("A sub group")
    color = gds.ColorItem("Color", default="red")
    choice1 = gds.ChoiceItem(
        "Single choice 1",
        [("16", "first choice"), ("32", "second choice"), ("64", "third choice")],
    )
    choice2 = gds.ChoiceItem("Image type", ImageTypes, default=ImageTypes.GAUSS)

    # LabeledEnum examples - Pattern 1 with invariant keys and UI labels
    filter_method = gds.ChoiceItem(
        "Filtering method", FilteringMethod, default=FilteringMethod.GAUSSIAN
    )
    threshold_mode = gds.ChoiceItem(
        "Threshold mode", ThresholdMode, default=ThresholdMode.OTSU
    )

    mchoice2 = gds.ImageChoiceItem(
        "Single choice 2",
        [
            ("rect", "first choice", "gif.png"),
            ("ell", "second choice", "txt.png"),
            ("qcq", "third choice", "file.png"),
        ],
    )
    _eg = gds.EndGroup("A sub group")
    floatarray = gds.FloatArrayItem(
        "Float array", default=np.ones((50, 5), float), format=" %.2e "
    ).set_pos(col=1)
    mchoice3 = gds.MultipleChoiceItem(
        "MC type 1", [str(i) for i in range(12)]
    ).horizontal(4)
    mchoice1 = (
        gds.MultipleChoiceItem(
            "MC type 2", ["first choice", "second choice", "third choice"]
        )
        .vertical(1)
        .set_pos(col=1)
    )
    dictionary = gds.DictItem(
        "Dictionary",
        default={
            "lkl": 2,
            "tototo": 3,
            "zzzz": "lklk",
            "bool": True,
            "float": 1.234,
            "list": [1, 2.5, 3, "str", False, 5, {"lkl": 2, "l": [1, 2, 3]}],
        },
        help="This is a dictionary",
    )


def test_all_items():
    """Test all DataItem objects"""
    with qt_app_context():
        prm = Parameters()

        prm.floatarray[:, 0] = np.linspace(-5, 5, 50)

        # Demonstrate LabeledEnum functionality
        execenv.print("=== LabeledEnum Pattern 1 Demo ===")
        execenv.print(f"Filter method (enum member): {prm.filter_method!r}")
        execenv.print(f"Filter value (invariant key): {prm.filter_method.value!r}")
        execenv.print(f"Filter label (UI display): {prm.filter_method.label!r}")
        execenv.print(f"Filter str(): {str(prm.filter_method)!r}")

        # Show different ways to set LabeledEnum values
        prm.filter_method = "sobel"  # Set by invariant key
        execenv.print(f"Set by key 'sobel': {prm.filter_method!r}")

        prm.filter_method = "Median filter"  # Set by UI label
        execenv.print(f"Set by label 'Median filter': {prm.filter_method!r}")

        prm.filter_method = 0  # Set by index
        execenv.print(f"Set by index 0: {prm.filter_method!r}")

        execenv.print("=== End LabeledEnum Demo ===")

        execenv.print(prm)
        if prm.edit():
            execenv.print(prm)
        prm.view()
        execenv.print("OK")


if __name__ == "__main__":
    test_all_items()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""Unit tests for DataSet item order in inheritance"""

import unittest

import guidata.dataset as gds


class DatasetA(gds.DataSet):
    """Dataset A with two items"""

    a1 = gds.FloatItem("a1")
    a2 = gds.IntItem("a2")


class DatasetB(gds.DataSet):
    """Dataset B with two items"""

    b1 = gds.TextItem("b1")
    b2 = gds.BoolItem("b2")


class DerivedAB(DatasetA, DatasetB):
    """Derived dataset from A and B"""

    d = gds.FloatItem("d")


class DerivedBA(DatasetB, DatasetA):
    """Derived dataset from B and A"""

    d = gds.FloatItem("d")


class DerivedSimple(DatasetA):
    """Derived dataset from DatasetA with an additional item"""

    d = gds.IntItem("d")


class TestDataSetItemOrder(unittest.TestCase):
    """Test DataSet item order in inheritance"""

    def assertItemNames(
        self, dataset_cls: type[gds.DataSet], expected_names: list[str]
    ) -> None:
        """Assert that the item names in the dataset match the expected names"""
        instance = dataset_cls()
        actual_names = [item.get_name() for item in instance.get_items()]
        self.assertEqual(actual_names, expected_names)

    def test_single_dataset(self) -> None:
        """Test single dataset item order"""
        self.assertItemNames(DatasetA, ["a1", "a2"])

    def test_simple_inheritance(self) -> None:
        """Test simple inheritance with one additional item"""
        self.assertItemNames(DerivedSimple, ["a1", "a2", "d"])

    def test_multiple_inheritance_AB(self) -> None:
        """Test multiple inheritance with DatasetA and DatasetB"""
        self.assertItemNames(DerivedAB, ["a1", "a2", "b1", "b2", "d"])

    def test_multiple_inheritance_BA(self) -> None:
        """Test multiple inheritance with DatasetB and DatasetA"""
        self.assertItemNames(DerivedBA, ["b1", "b2", "a1", "a2", "d"])

    def test_original_test_case(self) -> None:
        """Test original test case with double inheritance"""

        class OriginalDataset1(gds.DataSet):
            text1 = gds.TextItem("Text 1")
            int1 = gds.IntItem("Integer 1")

        class OriginalDataset2(gds.DataSet):
            text2 = gds.TextItem("Text 2")
            int2 = gds.IntItem("Integer 2")

        class DoubleInheritedDataset1(OriginalDataset1, OriginalDataset2):
            text3 = gds.TextItem("Text 3")
            int3 = gds.IntItem("Integer 3")

        class DoubleInheritedDataset2(OriginalDataset2, OriginalDataset1):
            text4 = gds.TextItem("Text 4")
            int4 = gds.IntItem("Integer 4")

        self.assertItemNames(
            DoubleInheritedDataset1, ["text1", "int1", "text2", "int2", "text3", "int3"]
        )
        self.assertItemNames(
            DoubleInheritedDataset2, ["text2", "int2", "text1", "int1", "text4", "int4"]
        )


if __name__ == "__main__":
    unittest.main()



#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Data items
----------

Base class
^^^^^^^^^^

.. autoclass:: guidata.dataset.DataItem

Numeric items
^^^^^^^^^^^^^

.. autoclass:: guidata.dataset.FloatItem
    :members:

.. autoclass:: guidata.dataset.IntItem
    :members:

.. autoclass:: guidata.dataset.FloatArrayItem
    :members:

Text items
^^^^^^^^^^

.. autoclass:: guidata.dataset.StringItem
    :members:

.. autoclass:: guidata.dataset.TextItem
    :members:

Date and time items
^^^^^^^^^^^^^^^^^^^

.. autoclass:: guidata.dataset.DateItem
    :members:

.. autoclass:: guidata.dataset.DateTimeItem
    :members:

Color items
^^^^^^^^^^^

.. autoclass:: guidata.dataset.ColorItem
    :members:

File items
^^^^^^^^^^

.. autoclass:: guidata.dataset.FileSaveItem
    :members:

.. autoclass:: guidata.dataset.FileOpenItem
    :members:

.. autoclass:: guidata.dataset.FilesOpenItem
    :members:

.. autoclass:: guidata.dataset.DirectoryItem
    :members:

Choice items
^^^^^^^^^^^^

.. autoclass:: guidata.dataset.BoolItem
    :members:

.. autoclass:: guidata.dataset.ChoiceItem
    :members:

.. autoclass:: guidata.dataset.MultipleChoiceItem
    :members:

.. autoclass:: guidata.dataset.ImageChoiceItem
    :members:

Other items
^^^^^^^^^^^

.. autoclass:: guidata.dataset.ButtonItem
    :members:

.. autoclass:: guidata.dataset.DictItem
    :members:

.. autoclass:: guidata.dataset.FontFamilyItem
    :members:
"""

from __future__ import annotations

import datetime
import os
import re
from collections.abc import Callable, Sequence
from enum import Enum, EnumMeta
from typing import TYPE_CHECKING, Any, Generic, Iterable, TypeVar

import numpy as np

from guidata.config import _
from guidata.dataset.datatypes import DataItem, DataSet, ItemProperty

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from guidata.io import (
        HDF5Reader,
        HDF5Writer,
        INIReader,
        INIWriter,
        JSONReader,
        JSONWriter,
    )

_T = TypeVar("_T")


class LabeledEnum(str, Enum):
    """Enum with invariant key and optional translated label.

    This enum supports Pattern 1:
    - .value is the invariant key (used for API, serialization, comparisons)
    - .label is the human-readable string (translated or not) used for UI display
    - Seamless interoperability: enum_member == string_value works

    Example:
        class MyEnum(LabeledEnum):
            OPTION1 = "opt1", _("Option 1")
            OPTION2 = "opt2", _("Option 2")
            OPTION3 = "opt3"  # Uses key as label

        # These are equivalent:
        MyEnum.OPTION1 == "opt1"  # True
        func(MyEnum.OPTION1) == func("opt1")  # Same behavior
    """

    def __new__(cls, value, label=None):
        if label is None:
            label = value
        # Initialize the str part with the value for consistent string behavior
        # This ensures string operations, comparisons, and isinstance(obj, str) work
        obj = str.__new__(cls, value)
        obj._value_ = value  # stable key
        obj.label = label  # UI label (possibly translated)
        return obj

    def __str__(self) -> str:
        """Return the label for display purposes."""
        return str(self.label)

    def __repr__(self) -> str:
        """Return the string representation of the enum value.

        This is crucial for code generation where enum members are used as default
        parameter values. By returning the string value directly, we avoid the need
        for makefun to resolve enum class references in generated function signatures.
        """
        return repr(self._value_)

    def __format__(self, format_spec):
        """Use the label when formatting for display."""
        return format(str(self.label), format_spec)

    def __eq__(self, other) -> bool:
        """Enable seamless comparison between enum members and their string values."""
        if isinstance(other, self.__class__):
            return self._value_ == other._value_
        elif isinstance(other, str):
            return self._value_ == other
        return False

    def __hash__(self) -> int:
        """Use the value for hashing to enable set operations with strings."""
        return hash(self._value_)


class NumericTypeItem(DataItem):
    """Numeric data item

    Args:
        label: item name
        default: default value (optional)
        min: minimum value (optional)
        max: maximum value (optional)
        nonzero: if True, zero is not a valid value (optional)
        unit: physical unit (optional)
        even: if True, even values are valid, if False,
          odd values are valid if None (default), ignored (optional)
        slider: if True, shows a slider widget right after the line
          edit widget (default is False)
        help: text shown in tooltip (optional)
        check: if False, value is not checked (optional, default=True)
        allow_none: if True, None is a valid value regardless of other constraints
          (optional, default=False)
    """

    type: type[int | float]

    def __init__(
        self,
        label: str,
        default: float | int | None = None,
        min: float | int | None = None,
        max: float | int | None = None,
        nonzero: bool | None = None,
        unit: str = "",
        help: str = "",
        check: bool = True,
        allow_none: bool = False,
    ) -> None:
        super().__init__(
            label, default=default, help=help, check=check, allow_none=allow_none
        )
        self.set_prop("data", min=min, max=max, nonzero=nonzero, check_value=check)
        self.set_prop("display", unit=unit)

    def get_auto_help(self, instance: DataSet) -> str:
        """Override DataItem method"""
        auto_help = {int: _("integer"), float: _("float")}[self.type]
        _min = self.get_prop_value("data", instance, "min")
        _max = self.get_prop_value("data", instance, "max")
        nonzero = self.get_prop_value("data", instance, "nonzero")
        unit = self.get_prop_value("display", instance, "unit")
        if _min is not None and _max is not None:
            auto_help += _(" between ") + str(_min) + _(" and ") + str(_max)
        elif _min is not None:
            auto_help += _(" higher than ") + str(_min)
        elif _max is not None:
            auto_help += _(" lower than ") + str(_max)
        if nonzero:
            auto_help += ", " + _("non zero")
        if unit:
            auto_help += ", %s %s" % (_("unit:"), unit)
        return auto_help

    def format_string(
        self, instance: DataSet, value: float | int, fmt: str, func: Callable
    ) -> str:
        """Override DataItem method"""
        text = fmt % (func(value),)
        # We add directly the unit to 'text' (instead of adding it
        # to 'fmt') to avoid string formatting error if '%' is in unit
        unit = self.get_prop_value("display", instance, "unit", "")
        if unit:
            text += " " + unit
        return text

    def check_value(self, value: float | int, raise_exception: bool = False) -> bool:
        """Override DataItem method"""
        if not self.get_prop("data", "check_value", True):
            return True
        if not isinstance(value, self.type):
            if raise_exception:
                raise TypeError(f"Expected {self.type}, got {type(value)}")
            return False
        if self.get_prop("data", "nonzero") and value == 0:
            if raise_exception:
                raise ValueError("Zero is not a valid value")
            return False
        _min = self.get_prop("data", "min")
        if _min is not None and value < _min:
            if raise_exception:
                raise ValueError(f"Value {value} is lower than minimum {_min}")
            return False
        _max = self.get_prop("data", "max")
        if _max is not None and value > _max:
            if raise_exception:
                raise ValueError(f"Value {value} is greater than maximum {_max}")
            return False
        return True

    def from_string(self, value: str) -> Any | None:
        """Override DataItem method"""
        # String may contains numerical operands:
        if re.match(r"^([\d\(\)\+/\-\*.]|e)+$", value):
            # pylint: disable=eval-used
            # pylint: disable=broad-except
            try:
                # pylint: disable=not-callable
                return self.type(eval(value))
            except:  # noqa
                pass
        return None


class FloatItem(NumericTypeItem):
    """Construct a float data item

    Args:
        label: item name
        default: default value (optional)
        min: minimum value (optional)
        max: maximum value (optional)
        nonzero: if True, zero is not a valid value (optional)
        unit: physical unit (optional)
        even: if True, even values are valid, if False,
         odd values are valid if None (default), ignored (optional)
        slider: if True, shows a slider widget right after the line
         edit widget (default is False)
        help: text shown in tooltip (optional)
        check: if False, value is not checked (optional, default=True)
        allow_none: if True, None is a valid value regardless of other constraints
         (optional, default=False)
    """

    type = float

    def __init__(
        self,
        label: str,
        default: float | None = None,
        min: float | None = None,
        max: float | None = None,
        nonzero: bool | None = None,
        unit: str = "",
        step: float = 0.1,
        slider: bool = False,
        help: str = "",
        check: bool = True,
        allow_none: bool = False,
    ) -> None:
        super().__init__(
            label,
            default=default,
            min=min,
            max=max,
            nonzero=nonzero,
            unit=unit,
            help=help,
            check=check,
            allow_none=allow_none,
        )
        self.set_prop("display", slider=slider)
        self.set_prop("data", step=step)

    def _set_value_with_validation(
        self, instance: Any, value: Any, force_allow_none: bool = False
    ) -> None:
        """Override DataItem._set_value_with_validation to convert integers to float"""
        # Try to convert NumPy numeric types to Python float
        # (will convert silently either floating point or integer types to float)
        try:
            if hasattr(value, "dtype") and not isinstance(value, self.type):
                value = self.type(value)
        except (TypeError, ValueError):
            pass
        # Now acceptable values could be either float or int
        # (no more NumPy types at this point)
        if isinstance(value, int):
            value = float(value)
        super()._set_value_with_validation(instance, value, force_allow_none)

    def get_value_from_reader(
        self, reader: HDF5Reader | JSONReader | INIReader
    ) -> float:
        """Reads value from the reader object, inside the try...except
        statement defined in the base item `deserialize` method"""
        return reader.read_float()


class IntItem(NumericTypeItem):
    """Construct an integer data item

    Args:
        label: item name
        default: default value (optional)
        min: minimum value (optional)
        max: maximum value (optional)
        nonzero: if True, zero is not a valid value (optional)
        unit: physical unit (optional)
        even: if True, even values are valid, if False,
         odd values are valid if None (default), ignored (optional)
        slider: if True, shows a slider widget right after the line
         edit widget (default is False)
        help: text shown in tooltip (optional)
        check: if False, value is not checked (optional, default=True)
        allow_none: if True, None is a valid value regardless of other constraints
         (optional, default=False)
    """

    type = int

    def __init__(
        self,
        label: str,
        default: int | None = None,
        min: int | None = None,
        max: int | None = None,
        nonzero: bool | None = None,
        unit: str = "",
        even: bool | None = None,
        slider: bool = False,
        help: str = "",
        check: bool = True,
        allow_none: bool = False,
    ) -> None:
        super().__init__(
            label,
            default=default,
            min=min,
            max=max,
            nonzero=nonzero,
            unit=unit,
            help=help,
            check=check,
            allow_none=allow_none,
        )
        self.set_prop("data", even=even)
        self.set_prop("display", slider=slider)

    def get_auto_help(self, instance: DataSet) -> str:
        """Override DataItem method"""
        auto_help = super().get_auto_help(instance)
        even = self.get_prop_value("data", instance, "even")
        if even is not None:
            if even:
                auto_help += ", " + _("even")
            else:
                auto_help += ", " + _("odd")
        return auto_help

    def _set_value_with_validation(
        self, instance: Any, value: Any, force_allow_none: bool = False
    ) -> None:
        """Override DataItem._set_value_with_validation
        to convert NumPy numeric types to int"""
        # Try to convert NumPy integer types to Python int
        # (will convert silently only integer types to int)
        try:
            if isinstance(value, np.integer):
                value = self.type(value)
        except (TypeError, ValueError):
            pass
        super()._set_value_with_validation(instance, value, force_allow_none)

    def check_value(self, value: int, raise_exception: bool = False) -> bool:
        """Override DataItem method"""
        if not self.get_prop("data", "check_value", True):
            return True
        valid = super().check_value(value, raise_exception=raise_exception)
        if not valid:
            return False
        even = self.get_prop("data", "even")
        if even is not None:
            is_even = value // 2 == value / 2.0
            if (even and not is_even) or (not even and is_even):
                if raise_exception:
                    oddity = "even" if even else "odd"
                    raise ValueError(f"Value {value} is not {oddity}")
                return False
        return True

    def get_value_from_reader(self, reader: HDF5Reader | JSONReader | INIReader) -> Any:
        """Reads value from the reader object, inside the try...except
        statement defined in the base item `deserialize` method"""
        return reader.read_int()


class StringItem(DataItem):
    """Construct a string data item

    Args:
        label: item name
        default: default value (optional)
        notempty: if True, empty string is not a valid value (optional)
        wordwrap: toggle word wrapping (optional)
        password: if True, text is hidden (optional)
        regexp: regular expression for checking value (optional)
        help: text shown in tooltip (optional)
        check: if False, value is not checked (ineffective for strings)
        allow_none: if True, None is a valid value regardless of other constraints
         (optional, default=False)
        readonly: if True, the item is read-only (optional, default=False)
    """

    type: Any = str

    def __init__(
        self,
        label: str,
        default: str | None = None,
        notempty: bool | None = None,
        wordwrap: bool = False,
        password: bool = False,
        regexp: str | None = None,
        help: str = "",
        check: bool = True,
        allow_none: bool = False,
        readonly: bool = False,
    ) -> None:
        super().__init__(
            label, default=default, help=help, check=check, allow_none=allow_none
        )
        self.set_prop("data", notempty=notempty, regexp=regexp)
        self.set_prop(
            "display", wordwrap=wordwrap, password=password, readonly=readonly
        )

    def get_auto_help(self, instance: DataSet) -> str:
        """Override DataItem method"""
        auto_help = _("string")
        notempty = self.get_prop_value("data", instance, "notempty")
        if notempty:
            auto_help += ", " + _("not empty")
        regexp = self.get_prop_value("data", instance, "regexp")
        if regexp:
            # Wrap regexp in backticks to prevent Sphinx from interpreting
            # it as a reference
            auto_help += ", " + _("regexp:") + f" ``{regexp}``"
        return auto_help

    def check_value(self, value: str, raise_exception: bool = False) -> bool:
        """Override DataItem method"""
        if not self.get_prop("data", "check_value", True):
            return True

        # Convert LabeledEnum members to their string values for validation
        if isinstance(value, LabeledEnum):
            value = value.value

        if not isinstance(value, self.type):
            if raise_exception:
                raise TypeError(f"Expected {self.type}, got {type(value)}")
            return False

        # Get all data properties at once to reduce property lookups
        data_props = self._props.get("data", {})

        notempty = data_props.get("notempty")
        if notempty and not value:
            if raise_exception:
                raise ValueError("Empty string is not a valid value")
            return False

        regexp = data_props.get("regexp")
        if regexp is not None:
            ok = bool(re.match(regexp, value or ""))
            if not ok and raise_exception:
                raise ValueError(f"Value {value} does not match regexp {regexp}")
            return ok
        return True

    def from_string(self, value: str) -> str:
        """Override DataItem method"""
        return value

    def get_string_value(self, instance: DataSet) -> str:
        """Override DataItem method"""
        strval = super().get_string_value(instance)
        if self.get_prop("display", "password"):
            return "*" * len(strval)
        return strval

    def _set_value_with_validation(
        self, instance: Any, value: Any, force_allow_none: bool = False
    ) -> None:
        """Override DataItem._set_value_with_validation to handle LabeledEnum members"""
        # Convert LabeledEnum members to their string values before validation
        if isinstance(value, LabeledEnum):
            value = value.value
        super()._set_value_with_validation(instance, value, force_allow_none)

    def get_value_from_reader(self, reader: HDF5Reader | JSONReader | INIReader) -> Any:
        """Reads value from the reader object, inside the try...except
        statement defined in the base item `deserialize` method"""
        return reader.read_unicode()


class TextItem(StringItem):
    """Construct a text data item (multiline string)

    Args:
        label: item name
        default: default value (optional)
        notempty: if True, empty string is not a valid value (optional)
        wordwrap: toggle word wrapping (optional)
        help: text shown in tooltip (optional)
        allow_none: if True, None is a valid value regardless of other constraints
         (optional, default=False)
        readonly: if True, the item is read-only (optional, default=False)
        regexp: regular expression for value validation (optional)
    """

    def __init__(
        self,
        label: str,
        default: str | None = None,
        notempty: bool | None = None,
        wordwrap: bool = True,
        help: str = "",
        allow_none: bool = False,
        readonly: bool = False,
        regexp: str | None = None,
    ) -> None:
        super().__init__(
            label,
            default=default,
            notempty=notempty,
            wordwrap=wordwrap,
            help=help,
            allow_none=allow_none,
            readonly=readonly,
            regexp=regexp,
        )


class BoolItem(DataItem):
    """Construct a boolean data item

    Args:
        text: form's field name (optional)
        label: item name
        default: default value (optional)
        help: text shown in tooltip (optional)
        check: if False, value is not checked (optional, default=True)
        allow_none: if True, None is a valid value regardless of other constraints
         (optional, default=False)
    """

    type = bool

    def __init__(
        self,
        text: str = "",
        label: str = "",
        default: bool | None = None,
        help: str = "",
        check: bool = True,
        allow_none: bool = False,
    ) -> None:
        super().__init__(
            label, default=default, help=help, check=check, allow_none=allow_none
        )
        self.set_prop("display", text=text)

    def get_string_value(self, instance: DataSet) -> str:
        """Override DataItem method"""
        value_str = "☑" if self.get_value(instance) else "☐"
        label = self.get_prop_value("display", self, "label")
        text = self.get_prop_value("display", instance, "text", "")
        if label and text:
            value_str += " " + text
        return value_str

    def get_value_from_reader(
        self, reader: HDF5Reader | JSONReader | INIReader
    ) -> bool:
        """Reads value from the reader object, inside the try...except
        statement defined in the base item `deserialize` method"""
        return reader.read_bool()


class DateItem(DataItem):
    """DataSet data item

    Args:
        label: item label
        default: default value (optional)
        format: date format (as in :py:func:`datetime.date.strftime`)
        help: text displayed on data item's tooltip
        check: check value (default: True)
        allow_none: if True, None is a valid value regardless of other constraints
         (optional, default=False)
    """

    type = datetime.date

    def __init__(
        self,
        label: str,
        default: datetime.date | None = None,
        format: str | None = None,
        help: str | None = "",
        check: bool | None = True,
        allow_none: bool = False,
    ) -> None:
        super().__init__(
            label, default=default, help=help, check=check, allow_none=allow_none
        )
        self.set_prop("display", format=format)


class DateTimeItem(DateItem):
    """DataSet data item

    Args:
        label: item label
        default: default value (optional)
        format: date format (as in :py:func:`datetime.date.strftime`)
        help: text displayed on data item's tooltip
        check: check value (default: True)
        allow_none: if True, None is a valid value regardless of other constraints
         (optional, default=False)
    """

    type = datetime.datetime

    def __init__(
        self,
        label: str,
        default: datetime.datetime | None = None,
        format: str | None = None,
        help: str | None = "",
        check: bool | None = True,
        allow_none: bool = False,
    ) -> None:
        super().__init__(label, default, format, help, check, allow_none=allow_none)


class ColorItem(StringItem):
    """Construct a color data item

    Args:
        label: item name
        default: default value (optional)
        notempty: if True, empty string is not a valid value (optional)
        regexp: regular expression for checking value (optional)
        help: text shown in tooltip (optional)
        check: if False, value is not checked (ineffective for strings)
        allow_none: if True, None is a valid value regardless of other constraints
         (optional, default=True)
    """

    def __init__(
        self,
        label: str,
        default: str | None = None,
        notempty: bool | None = None,
        regexp: str | None = None,
        help: str = "",
        check: bool = True,
        allow_none: bool = True,
    ) -> None:
        super().__init__(
            label,
            default=default,
            notempty=notempty,
            regexp=regexp,
            help=help,
            check=check,
            allow_none=allow_none,
        )

    def check_value(self, value: str, raise_exception: bool = False) -> bool:
        """Override DataItem method"""
        if not self.get_prop("data", "check_value", True):
            return True
        if not isinstance(value, self.type):
            if raise_exception:
                raise TypeError(f"Expected {self.type}, got {type(value)}")
            return False
        from qtpy import QtGui as QG

        ok = QG.QColor(value).isValid()
        if not ok and raise_exception:
            raise ValueError(f"Value {value} is not a valid color")
        return ok

    def get_value_from_reader(self, reader: HDF5Reader | JSONReader | INIReader) -> str:
        """Reads value from the reader object, inside the try...except
        statement defined in the base item `deserialize` method"""
        # Using read_str converts `numpy.string_` to `str` -- otherwise,
        # when passing the string to a QColor Qt object, any numpy.string_ will
        # be interpreted as no color (black)
        return reader.read_str()


class FileSaveItem(StringItem):
    """Construct a path data item for a file to be saved

    Args:
        label: item name
        formats: wildcard filter
        default: default value (optional)
        basedir: default base directory (optional)
        regexp: regular expression for checking value (optional)
        help: text shown in tooltip (optional)
        check: if False, value is not checked (optional, default=True)
        all_files_first: if True, "All files" is the first item in the list
         (optional, default=False)
    """

    def __init__(
        self,
        label: str,
        formats: tuple[str, ...] | str = "*",
        default: list[str] | str | None = None,
        basedir: str | None = None,
        all_files_first: bool = False,
        regexp: str | None = None,
        help: str = "",
        check: bool = True,
        allow_none: bool = False,
    ) -> None:
        default = os.path.join(*default) if isinstance(default, list) else default

        super().__init__(
            label,
            default=default,
            regexp=regexp,
            help=help,
            check=check,
            allow_none=allow_none,
        )
        if isinstance(formats, str):
            formats = [formats]  # type:ignore
        self.set_prop("data", formats=formats)
        self.set_prop("data", basedir=basedir)
        self.set_prop("data", all_files_first=all_files_first)
        self.set_prop("display", func=os.path.basename)

    def get_auto_help(self, instance: DataSet) -> str:
        """Override DataItem method"""
        formats = self.get_prop("data", "formats")
        return (
            _("all file types")
            if formats == ["*"]
            else _("supported file types:") + " *.%s" % ", *.".join(formats)
        )

    def check_value(self, value: str, raise_exception: bool = False) -> bool:
        """Override DataItem method"""
        if not self.get_prop("data", "check_value", True):
            return True
        if not isinstance(value, self.type):
            if raise_exception:
                raise TypeError(f"Expected {self.type}, got {type(value)}")
            return False
        ok = len(value) > 0
        if not ok and raise_exception:
            raise ValueError("Empty string is not a valid value")
        return ok

    def from_string(self, value) -> str:
        """Override DataItem method"""
        return self.add_extension(value)

    def add_extension(self, value) -> str:
        """Add extension to filename
        `value`: possible value for data item"""
        value = str(value)
        formats = self.get_prop("data", "formats")
        if (
            len(formats) == 1
            and formats[0] != "*"
            and not value.endswith("." + formats[0])
            and len(value) > 0
        ):
            return value + "." + formats[0]
        return value


class FileOpenItem(FileSaveItem):
    """Construct a path data item for a file to be opened

    Args:
        label: item name
        formats: wildcard filter
        default: default value (optional)
        basedir: default base directory (optional)
        help: text shown in tooltip (optional)
        check: if False, value is not checked (optional, default=True)
    """

    def check_value(self, value: str, raise_exception: bool = False) -> bool:
        """Override DataItem method"""
        if not self.get_prop("data", "check_value", True):
            return True
        if not isinstance(value, self.type):
            if raise_exception:
                raise TypeError(f"Expected {self.type}, got {type(value)}")
            return False
        ok = os.path.exists(value) and os.path.isfile(value)
        if not ok and raise_exception:
            raise ValueError(f"File {value} does not exist or is not a file")
        return ok


class FilesOpenItem(FileSaveItem):
    """Construct a path data item for multiple files to be opened.

    Args:
        label: item name
        formats: wildcard filter
        default: default value (optional)
        basedir: default base directory (optional)
        regexp: regular expression for checking value (optional)
        help: text shown in tooltip (optional)
        check: if False, value is not checked (optional, default=True)
        all_files_first: if True, "All files" is the first item in the list
         (optional, default=False)
    """

    type = list

    def __init__(
        self,
        label: str,
        formats: str = "*",
        default: list[str] | str | None = None,
        basedir: str | None = None,
        all_files_first: bool = False,
        regexp: str | None = None,
        help: str = "",
        check: bool = True,
        allow_none: bool = False,
    ) -> None:
        if isinstance(default, str):
            default = [default]
        StringItem.__init__(
            self,
            label,
            default=default,
            regexp=regexp,
            help=help,
            check=check,
            allow_none=allow_none,
        )
        if isinstance(formats, str):
            formats = [formats]  # type:ignore
        self.set_prop("data", formats=formats)
        self.set_prop("data", basedir=basedir)
        self.set_prop("data", all_files_first=all_files_first)
        self.set_prop("display", func=self.paths_basename)

    @staticmethod
    def paths_basename(paths: str | list[str]):
        """Return the basename of a path or a list of paths"""
        return (
            [os.path.basename(p) for p in paths]
            if isinstance(paths, list)
            else os.path.basename(paths)
        )

    def check_value(self, value: str, raise_exception: bool = False) -> bool:
        """Override DataItem method"""
        if not self.get_prop("data", "check_value", True):
            return True
        if value is None:
            if raise_exception:
                raise ValueError("Value cannot be None")
            return False
        allexist = True
        for path in value:
            allexist = allexist and os.path.exists(path) and os.path.isfile(path)
        if not allexist and raise_exception:
            raise ValueError(f"Some files do not exist or are not files: {value}")
        return allexist

    def from_string(self, value: Any) -> list[str]:  # type:ignore
        """Override DataItem method"""
        value = eval(value) if value.endswith("']") or value.endswith('"]') else [value]
        return [self.add_extension(path) for path in value]

    def serialize(
        self,
        instance: DataSet,
        writer: HDF5Writer | JSONWriter | INIWriter,
    ) -> None:
        """Serialize this item"""
        value = self.get_value(instance)
        if value is not None and not isinstance(value, (tuple, list)):
            value = [value]
        writer.write_sequence([fname.encode("utf-8") for fname in value])

    def get_value_from_reader(
        self, reader: HDF5Reader | JSONReader | INIReader
    ) -> list[str]:
        """Reads value from the reader object, inside the try...except
        statement defined in the base item `deserialize` method"""
        return [fname for fname in reader.read_sequence()]


class DirectoryItem(StringItem):
    """Construct a path data item for a directory.

    Args:
        label: item name
        default: default value (optional)
        help: text shown in tooltip (optional)
        check: if False, value is not checked (optional, default=True)
    """

    def check_value(self, value: str, raise_exception: bool = False) -> bool:
        """Override DataItem method"""
        if not self.get_prop("data", "check_value", True):
            return True
        if not isinstance(value, self.type):
            if raise_exception:
                raise TypeError(f"Expected {self.type}, got {type(value)}")
            return False
        ok = os.path.exists(value) and os.path.isdir(value)
        if not ok and raise_exception:
            raise ValueError(f"Directory {value} does not exist or is not a directory")
        return ok


class FirstChoice:
    """
    Special object that means the default value of a ChoiceItem
    is the first item.
    """

    pass


class ChoiceItem(DataItem, Generic[_T]):
    """Construct a data item for a list of choices.

    Args:
        label: item name
        choices: string list or (key, label) list
         function of two arguments (item, value) returning a list of tuples
         (key, label, image) where image is an icon path, a QIcon instance
         or a function of one argument (key) returning a QIcon instance
        default: default value (optional)
        help: text shown in tooltip (optional)
        check: if False, value is not checked (optional, default=True)
        radio: if True, shows radio buttons instead of a combo box
         (default is False)
        size: size (optional) of the combo box or button widget (for radio buttons)
        allow_none: if True, None is a valid value regardless of other constraints
         (optional, default=True)
    """

    type = Any

    def __init__(
        self,
        label: str,
        choices: Iterable[_T] | Callable[[Any], Iterable[_T]] | EnumMeta,
        default: tuple[()] | type[FirstChoice] | int | _T | Enum | None = FirstChoice,
        help: str = "",
        check: bool = True,
        radio: bool = False,
        size: tuple[int, int] | None = None,
        allow_none: bool = True,
    ) -> None:
        self._enum_cls: type[Enum] | None = None
        _choices_data: Any
        if isinstance(choices, EnumMeta):
            self._enum_cls = choices
            # Build _choices_data as [(m.name, m.label, None)] for each member
            _choices_data = [
                (m.name, getattr(m, "label", str(m.value)), None) for m in choices
            ]
            # Only coerce if default is a real value
            if default not in (FirstChoice, None):
                default = self._enum_coerce_in(default)
        elif isinstance(choices, Callable):
            _choices_data = ItemProperty(choices)
        else:
            _choices_data = []
            for idx, choice in enumerate(choices):
                _choices_data.append(self._normalize_choice(idx, choice))
        if (
            default is FirstChoice
            and isinstance(_choices_data, Sequence)
            and isinstance(_choices_data[0], Sequence)
        ):
            default = _choices_data[0][0]
        elif default is FirstChoice:
            default = None
        super().__init__(
            label, default=default, help=help, check=check, allow_none=allow_none
        )
        self.set_prop("data", choices=_choices_data)
        self.set_prop("display", radio=radio)
        self.set_prop("display", size=size)

    def check_value(self, value: str, raise_exception: bool = False) -> bool:
        """Override DataItem method"""
        if not self.get_prop("data", "check_value", True):
            return True
        choices = self.get_prop("data", "choices", [])
        if not isinstance(choices, ItemProperty):
            values = [v for v, _k, _i in choices]
            # Check if value is in values, with special handling for sequences
            # (JSON serialization converts tuples to lists, so we need to compare
            # their content rather than using strict equality)
            value_found = False
            for v in values:
                if v == value:
                    value_found = True
                    break
                # If both are sequences (list/tuple), compare their content
                if isinstance(v, (list, tuple)) and isinstance(value, (list, tuple)):
                    if len(v) == len(value) and all(a == b for a, b in zip(v, value)):
                        value_found = True
                        break
            if not value_found:
                if raise_exception:
                    raise ValueError(
                        f"Invalid value '{value}' (valid values: {values})"
                    )
                return False
        return True

    def _enum_coerce_in(self, v: object) -> str:
        """Accept Enum member | name | value | label | index
        → return string name (the storage key).
        """
        if self._enum_cls is None:
            raise TypeError("No Enum class set in ChoiceItem")

        # Enum member
        if isinstance(v, self._enum_cls):
            return v.name

        # name (e.g. "LINEAR")
        if isinstance(v, str) and v in self._enum_cls.__members__:
            return v

        # value (stable key) or label (UI string)
        for m in self._enum_cls:
            if v == m.value or v == getattr(m, "label", str(m.value)):
                return m.name

        # index
        if isinstance(v, int):
            members = list(self._enum_cls)
            if 0 <= v < len(members):
                return members[v].name

        raise ValueError(
            f"Invalid value '{v}' for {self._enum_cls.__name__} "
            f"(expected a member, name, value, label, or index)"
        )

    def __get__(self, instance: DataSet, owner: type | None = None) -> Any:
        """Override DataItem.__get__ to return Enum member if applicable"""
        # descriptor for user access → return Enum member if applicable
        if instance is None:
            return self
        raw = super().__get__(instance, owner)  # stored key (string) from DataItem
        if self._enum_cls is not None and raw is not None:
            return self._enum_cls[raw]  # Enum member
        return raw  # legacy: keep as-is

    def get_value(self, instance: DataSet) -> str | None:
        """Override DataItem.get_value"""
        # guidata internals should call this; keep it returning the raw key
        return super().__get__(instance, instance.__class__)  # string (or None)

    def _set_value_with_validation(
        self, instance: Any, value: Any, force_allow_none: bool = False
    ) -> None:
        """Override DataItem._set_value_with_validation to accept Enum members"""
        if self._enum_cls is not None and value is not None:
            value = self._enum_coerce_in(value)  # → member.name
        super()._set_value_with_validation(instance, value, force_allow_none)

    def _normalize_choice(
        self, idx: int, choice_tuple: tuple[Any, ...]
    ) -> tuple[int, str, None] | tuple[str, str, None]:
        if isinstance(choice_tuple, tuple):
            key, value = choice_tuple
        else:
            key = idx
            value = choice_tuple
        return (key, value, None)

    def get_string_value(self, instance: DataSet) -> str:
        """Override DataItem method"""
        value = self.get_value(instance)
        choices = self.get_prop_value("data", instance, "choices")
        # print "ShowChoiceWidget:", choices, value
        for choice in choices:
            if choice[0] == value:
                return str(choice[1])
        return DataItem.get_string_value(self, instance)


class MultipleChoiceItem(ChoiceItem):
    """Construct a data item for a list of choices -- multiple choices can be selected

    Args:
        label: item name
        choices: string list or (key, label) list
        default: default value (optional)
        help: text shown in tooltip (optional)
        check: if False, value is not checked (optional, default=True)
        allow_none: if True, None is a valid value regardless of other constraints
         (optional, default=True)
    """

    def __init__(
        self,
        label: str,
        choices: list[str],
        default: tuple[()] = (),
        help: str = "",
        check: bool = True,
        allow_none: bool = True,
    ) -> None:
        super().__init__(
            label, choices, default, help, check=check, allow_none=allow_none
        )
        self.set_prop("display", shape=(1, -1))

    def check_value(self, value: str, raise_exception: bool = False) -> bool:
        """Override DataItem method"""
        if not self.get_prop("data", "check_value", True):
            return True
        if not isinstance(value, tuple):
            if raise_exception:
                raise ValueError(f"Invalid value '{value}' (expecting tuple)")
            return False
        for val in value:
            if val not in [v for v, _k, _i in self.get_prop("data", "choices", [])]:
                if raise_exception:
                    raise ValueError(f"Invalid value '{value}'")
                return False
        return True

    def horizontal(self, row_nb: int = 1) -> MultipleChoiceItem:
        """
        Method to arange choice list horizontally on `n` rows

        Example:
        nb = MultipleChoiceItem("Number", ['1', '2', '3'] ).horizontal(2)
        """
        self.set_prop("display", shape=(row_nb, -1))
        return self

    def vertical(self, col_nb: int = 1) -> MultipleChoiceItem:
        """
        Method to arange choice list vertically on `n` columns

        Example:
        nb = MultipleChoiceItem("Number", ['1', '2', '3'] ).vertical(2)
        """
        self.set_prop("display", shape=(-1, col_nb))
        return self

    def serialize(
        self,
        instance: DataSet,
        writer: HDF5Writer | JSONWriter | INIWriter,
    ) -> None:
        """Serialize this item"""
        value = self.get_value(instance)
        seq = []
        _choices = self.get_prop_value("data", instance, "choices")
        for key, _label, _img in _choices:
            seq.append(key in value)
        writer.write_sequence(seq)

    def deserialize(
        self,
        instance: DataSet,
        reader: HDF5Reader | JSONReader | INIReader,
    ) -> None:
        """Deserialize this item"""
        try:
            flags = reader.read_sequence()
        except KeyError:
            self.set_default(instance)
        else:
            # We could have trouble with objects providing their own choice
            # function which depend on not yet deserialized values
            _choices = self.get_prop_value("data", instance, "choices")
            value = []
            for idx, flag in enumerate(flags):
                if flag:
                    value.append(_choices[idx][0])
            self.__set__(instance, value)


class ImageChoiceItem(ChoiceItem):
    """Construct a data item for a list of choices with images

    Args:
        label: item name
        choices: (label, image) list, or (key, label, image) list,
         or function of two arguments (item, value) returning a list of tuples
         (key, label, image) where image is an icon path, a QIcon instance or
         a function of one argument (key) returning a QIcon instance
        default: default value (optional)
        help: text shown in tooltip (optional)
        radio: if True, shows radio buttons instead of a combo box
         (default is False)
        size: size (optional) of the combo box or button widget (for radio buttons)
        allow_none: if True, None is a valid value regardless of other constraints
         (optional, default=True)
    """

    def _normalize_choice(
        self, idx: int, choice_tuple: tuple[Any, ...]
    ) -> tuple[Any, Any, Any]:
        assert isinstance(choice_tuple, tuple)
        if len(choice_tuple) == 3:
            key, value, img = choice_tuple
        else:
            key = idx
            value, img = choice_tuple
        return (key, value, img)


class FloatArrayItem(DataItem):
    """Construct a float array data item

    Args:
        label: item name
        default: default value (optional)
        help: text shown in tooltip (optional)
        format: formatting string (example: '%.3f') (optional)
        transpose: transpose matrix (display only)
        large: view all float of the array
        minmax: "all" (default), "columns", "rows"
        check: if False, value is not checked (optional, default=True)
        variable_size: if True, allows to add/remove row/columns on all axis
        allow_none: if True, None is a valid value regardless of other constraints
         (optional, default=True)
        check_callback: additional callback to check the value
         (function of two arguments (value, raise_exception) returning a boolean,
         where value is the value to check and raise_exception is a boolean
         indicating whether to raise an exception on invalid value)
    """

    type = np.ndarray

    def __init__(
        self,
        label: str,
        default: NDArray | None = None,
        help: str = "",
        format: str = "%.3f",
        transpose: bool = False,
        minmax: str = "all",
        check: bool = True,
        variable_size=False,
        allow_none: bool = True,
        check_callback: Callable[[np.ndarray, bool], bool] | None = None,
    ) -> None:
        super().__init__(
            label, default=default, help=help, check=check, allow_none=allow_none
        )
        self.set_prop("display", format=format, transpose=transpose, minmax=minmax)
        self.set_prop("edit", variable_size=variable_size)
        self.check_callback = check_callback

    def check_value(self, value: np.ndarray, raise_exception: bool = False) -> bool:
        """Override DataItem method"""
        if not self.get_prop("data", "check_value", True):
            return True
        if not isinstance(value, self.type):
            if raise_exception:
                raise TypeError(f"Expected {self.type}, got {type(value)}")
            return False
        if self.check_callback is not None:
            return self.check_callback(value, raise_exception)
        return True

    def format_string(
        self, instance: DataSet, value: Any, fmt: str, func: Callable
    ) -> str:
        """Override DataItem method"""
        larg = self.get_prop_value("display", instance, "large", False)
        fmt = self.get_prop_value("display", instance, "format", "%s")
        unit = self.get_prop_value("display", instance, "unit", "")
        v: np.ndarray = func(value)
        if v.size == 0:
            return "= []"
        try:
            if larg:
                text = "= ["
                for flt in v[:-1]:
                    text += fmt % flt + "; "
                text += fmt % v[-1] + "]"
            else:
                text = "~= " + fmt % v.mean()
                text += " [" + fmt % v.min()
                text += " .. " + fmt % v.max()
                text += "]"
            text += " %s" % unit
            return str(text)
        except (ValueError, TypeError):
            return "= %s %s" % (str(value), unit)

    def serialize(
        self,
        instance: DataSet,
        writer: HDF5Writer | JSONWriter | INIWriter,
    ) -> None:
        """Serialize this item"""
        value = self.get_value(instance)
        writer.write_array(value)

    def get_value_from_reader(self, reader: HDF5Reader | JSONReader | INIReader) -> Any:
        """Reads value from the reader object, inside the try...except
        statement defined in the base item `deserialize` method"""
        return reader.read_array()


class DictItem(DataItem):
    """Construct a data item representing a dictionary

    Args:
        label: item name
        default: default value (optional)
        help: text shown in tooltip (optional)
        check: if False, value is not checked (optional, default=True)
        allow_none: if True, None is a valid value regardless of other constraints
         (optional, default=False)
    """

    type: type[dict[str, Any]] = dict

    # pylint: disable=redefined-builtin,abstract-method
    def __init__(
        self,
        label,
        default: dict | None = None,
        help="",
        check=True,
        allow_none: bool = False,
    ):
        super().__init__(
            label, default=default, help=help, check=check, allow_none=allow_none
        )
        self.set_prop("display", callback=self.__dictedit)
        self.set_prop("display", icon="dictedit.png")

    @staticmethod
    # pylint: disable=unused-argument
    def __dictedit(
        instance: DataSet,
        item: DataItem,
        value: dict,
        parent,
        trigger_apply=None,
    ):
        """Open a dictionary editor

        Args:
            instance: DataSet instance
            item: DataItem instance
            value: Current dictionary value
            parent: Parent widget
            trigger_apply: Optional callback to trigger auto-apply
        """
        # pylint: disable=import-outside-toplevel
        from guidata.qthelpers import exec_dialog
        from guidata.widgets.collectionseditor import CollectionsEditor

        editor = CollectionsEditor(parent)
        value_was_none = value is None
        if value_was_none:
            value = {}
        editor.setup(value, readonly=instance.is_readonly())
        result = exec_dialog(editor)
        if result:
            new_value = editor.get_value()
            # Auto-apply changes if trigger function was provided
            if trigger_apply is not None:
                trigger_apply()
            return new_value
        if value_was_none:
            return None
        return value

    def serialize(self, instance, writer):
        """Serialize this item"""
        value = self.get_value(instance)
        writer.write_dict(value)

    def get_value_from_reader(self, reader):
        """Reads value from the reader object, inside the try...except
        statement defined in the base item `deserialize` method"""
        return reader.read_dict()


class ButtonItem(DataItem):
    """Construct a simple button that calls a method when hit

    Args:
        label: item name
        callback: function with four parameters (dataset, item, value,
         parent) where dataset (DataSet) is an instance of the parent dataset,
         item (DataItem) is an instance of ButtonItem (i.e. self), value (unspecified)
         is the value of ButtonItem (default ButtonItem value or last value returned
         by the callback) and parent (QObject) is button's parent widget
        icon: icon show on the button (optional)
         (str: icon filename as in guidata/guiqwt image search paths)
        default: default value passed to the callback (optional)
        help: text shown in button's tooltip (optional)
        check: if False, value is not checked (optional, default=True)
        size: size (optional) of the button widget
        allow_none: if True, None is a valid value regardless of other constraints
         (optional, default=True)

    The value of this item is unspecified but is passed to the callback along
    with the whole dataset. The value is assigned the callback`s return value.
    """

    def __init__(
        self,
        label: str,
        callback: Callable,
        icon: str | None = None,
        default: Any | None = None,
        help: str = "",
        check: bool = True,
        size: tuple[int, int] | None = None,
        allow_none: bool = True,
    ) -> None:
        super().__init__(
            label, default=default, help=help, check=check, allow_none=allow_none
        )
        self.set_prop("display", callback=callback)
        self.set_prop("display", icon=icon)
        self.set_prop("display", size=size)

    def serialize(
        self,
        instance: DataSet,
        writer: HDF5Writer | JSONWriter | INIWriter,
    ) -> Any:
        pass

    def deserialize(
        self,
        instance: DataSet,
        reader: HDF5Reader | JSONReader | INIReader,
    ) -> Any:
        pass


class FontFamilyItem(StringItem):
    """Construct a font family name item

    Args:
        label: item name
        default: default value (optional)
        help: text shown in tooltip (optional)
        check: if False, value is not checked (optional, default=True)
    """

    pass



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Terminal emulation tools"""

import os


class ANSIEscapeCodeHandler(object):
    """ANSI Escape sequences handler"""

    if os.name == "nt":
        # Windows terminal colors:
        ANSI_COLORS = (  # Normal, Bright/Light
            ("#000000", "#808080"),  # 0: black
            ("#800000", "#ff0000"),  # 1: red
            ("#008000", "#00ff00"),  # 2: green
            ("#808000", "#ffff00"),  # 3: yellow
            ("#000080", "#0000ff"),  # 4: blue
            ("#800080", "#ff00ff"),  # 5: magenta
            ("#008080", "#00ffff"),  # 6: cyan
            ("#c0c0c0", "#ffffff"),  # 7: white
        )
    elif os.name == "mac":
        # Terminal.app colors:
        ANSI_COLORS = (  # Normal, Bright/Light
            ("#000000", "#818383"),  # 0: black
            ("#C23621", "#FC391F"),  # 1: red
            ("#25BC24", "#25BC24"),  # 2: green
            ("#ADAD27", "#EAEC23"),  # 3: yellow
            ("#492EE1", "#5833FF"),  # 4: blue
            ("#D338D3", "#F935F8"),  # 5: magenta
            ("#33BBC8", "#14F0F0"),  # 6: cyan
            ("#CBCCCD", "#E9EBEB"),  # 7: white
        )
    else:
        # xterm colors:
        ANSI_COLORS = (  # Normal, Bright/Light
            ("#000000", "#7F7F7F"),  # 0: black
            ("#CD0000", "#ff0000"),  # 1: red
            ("#00CD00", "#00ff00"),  # 2: green
            ("#CDCD00", "#ffff00"),  # 3: yellow
            ("#0000EE", "#5C5CFF"),  # 4: blue
            ("#CD00CD", "#ff00ff"),  # 5: magenta
            ("#00CDCD", "#00ffff"),  # 6: cyan
            ("#E5E5E5", "#ffffff"),  # 7: white
        )

    def __init__(self):
        self.intensity = 0
        self.italic = None
        self.bold = None
        self.underline = None
        self.foreground_color = None
        self.background_color = None
        self.default_foreground_color = 30
        self.default_background_color = 47

    def set_code(self, code):
        """

        :param code:
        """
        assert isinstance(code, int)
        if code == 0:
            # Reset all settings
            self.reset()
        elif code == 1:
            # Text color intensity
            self.intensity = 1
            # The following line is commented because most terminals won't
            # change the font weight, against ANSI standard recommendation:
        #            self.bold = True
        elif code == 3:
            # Italic on
            self.italic = True
        elif code == 4:
            # Underline simple
            self.underline = True
        elif code == 22:
            # Normal text color intensity
            self.intensity = 0
            self.bold = False
        elif code == 23:
            # No italic
            self.italic = False
        elif code == 24:
            # No underline
            self.underline = False
        elif code >= 30 and code <= 37:
            # Text color
            self.foreground_color = code
        elif code == 39:
            # Default text color
            self.foreground_color = self.default_foreground_color
        elif code >= 40 and code <= 47:
            # Background color
            self.background_color = code
        elif code == 49:
            # Default background color
            self.background_color = self.default_background_color
        self.set_style()

    def set_style(self):
        """
        Set font style with the following attributes:
        'foreground_color', 'background_color', 'italic',
        'bold' and 'underline'
        """
        raise NotImplementedError

    def reset(self):
        """ """
        self.current_format = None
        self.intensity = 0
        self.italic = False
        self.bold = False
        self.underline = False
        self.foreground_color = None
        self.background_color = None



# -*- coding: utf-8 -*-

"""
about
=====

"""

from __future__ import annotations

import platform
import sys

import qtpy
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QMainWindow, QMessageBox

import guidata
from guidata.config import _
from guidata.configtools import get_icon


def get_python_libs_infos(addinfo: str = "") -> str:
    """Get Python and libraries information

    Args:
        addinfo: additional information to be displayed

    Returns:
        str: Python and libraries information
    """
    python_version = "{} {}".format(
        platform.python_version(), "64 bits" if sys.maxsize > 2**32 else "32 bits"
    )
    if qtpy.PYQT_VERSION is None:
        qtb_version = qtpy.PYSIDE_VERSION
        qtb_name = "PySide"
    else:
        qtb_version = qtpy.PYQT_VERSION
        qtb_name = "PyQt"
    if addinfo:
        addinfo = ", " + addinfo
    return (
        f"Python {python_version}, "
        f"Qt {qtpy.QT_VERSION}, {qtb_name} {qtb_version}"
        f"{addinfo} on {platform.system()}"
    )


def get_general_infos(addinfo: str = "") -> str:
    """Get general information (copyright, Qt versions, etc.)

    Args:
        addinfo: additional information to be displayed

    Returns:
        str: Qt information
    """
    return "Copyright © 2023 CEA-Codra\n\n" + get_python_libs_infos(addinfo=addinfo)


class AboutInfo:
    """Object to generate information about the package

    Args:
        name: package name
        version: package version
        description: package description
        author: package author
        year: package year
        organization: package organization
        project_url: package project url
        doc_url: package documentation url
    """

    def __init__(
        self,
        name: str,
        version: str,
        description: str,
        author: str,
        year: int,
        organization: str,
        project_url: str = "",
        doc_url: str = "",
    ) -> None:
        self.name = name
        self.version = version
        self.description = description
        self.author = author
        self.year = year
        self.organization = organization
        if not project_url:
            project_url = f"https://github.com/PlotPyStack/{name}"
        self.project_url = project_url
        if not doc_url:
            doc_url = f"https://{name}.readthedocs.io"
        self.doc_url = doc_url

    def __str__(self) -> str:
        return self.about()

    def about(
        self, html: bool = True, copyright_only: bool = False, addinfo: str = ""
    ) -> str:
        """Return text about this package

        Args:
            html: return html text. Defaults to True.
            copyright_only: if True, return only copyright
            addinfo: additional information to be displayed

        Returns:
            Text about this package
        """
        auth, year, org = self.author, self.year, self.organization
        if html:
            author = f"<a href='https://github.com/{auth.replace(' ', '')}'>{auth}</a>"
            organization = f"<a href='https://github.com/{org}'>{org}</a>"
        shdesc = f"{self.name} {self.version}\n{self.description}"
        if html:
            shdesc += "\n\n"
            pname = _("Project website")
            dname = _("Documentation")
            plink = f"<a href='{self.project_url}'>{pname}</a>"
            dlink = f"<a href='{self.doc_url}'>{dname}</a>"
            shdesc += _("More details about %s on %s or %s") % (self.name, plink, dlink)
        shdesc += "\n\n" + _("Created by %s in %d") % (author, year) + "\n"
        shdesc += _("Maintained by the %s organization") % organization
        desc = get_general_infos(addinfo)
        if not copyright_only:
            desc = f"{shdesc}{desc}"
        if html:
            desc = desc.replace("\n", "<br />")
        return desc


def about(html: bool = True, copyright_only: bool = False) -> str:
    """Return text about this package

    Args:
        html: return html text. Defaults to True.
        copyright_only: if True, return only copyright

    Returns:
        Text about this package
    """
    info = AboutInfo(
        "guidata",
        guidata.__version__,
        _("Automatic GUI generation for easy dataset editing and display"),
        "Pierre Raybaut",
        2009,
        "PlotPyStack",
    )
    return info.about(html=html, copyright_only=copyright_only)


def show_about_dialog() -> None:
    """Show ``guidata`` about dialog"""
    win = QMainWindow(None)
    win.setAttribute(Qt.WA_DeleteOnClose)
    win.hide()
    win.setWindowIcon(get_icon("guidata.svg"))
    QMessageBox.about(win, _("About") + " guidata", about(html=True))
    win.close()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""Little translation test"""

# guitest: show

from guidata.config import _
from guidata.env import execenv

translations = (_("Some required entries are incorrect"),)


def test_translations():
    """Test translations"""
    for text in translations:
        execenv.print(text)
    execenv.print("OK")


if __name__ == "__main__":
    test_translations()
